#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Decidedly Object-Oriented Runtime-checking (DOOR) type hint inference
fixtures** (i.e., :mod:`pytest`-specific context managers passed as parameters
to unit tests exercising the :mod:`beartype.bite.infer_hint` function).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from pytest import fixture

# ....................{ FIXTURES                           }....................
@fixture(scope='session')
def bite_cases_infer_hint() -> (
    'Dict[beartype.BeartypeConf, Iterable[Tuple[object, object]]]'):
    '''
    Session-scoped fixture returning a dictionary mapping from each relevant
    beartype configuration to a corresponding iterable of all **type hint
    inference cases** specific to that configuration, efficiently cached across
    all tests requiring this fixture.

    This container is intentionally defined by the return of this fixture rather
    than as a global constant of this submodule. Why? Because the former safely
    defers all heavyweight imports required to define this container to the call
    of the first unit test requiring this fixture, whereas the latter unsafely
    performs those imports at pytest test collection time.

    Returns
    -------
    Dict[beartype.BeartypeConf, Iterable[Tuple[object, object]]]
        Dictionary mapping from each relevant beartype configuration to a
        corresponding iterable of all **type hint inference cases** specific to
        that configuration, where each type hint inference case is a 2-tuple
        ``(obj, hint)`` describing the type hint matching an arbitrary object
        as follows:

        * ``obj`` is an arbitrary object to be passed as the first parameter to
          the :func:`beartype.bite.infer_hint` function.
        * ``hint`` is the type hint returned by that function when passed that
          object.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer fixture-specific imports.
    from beartype import (
        BeartypeConf,
        BeartypeStrategy,
    )
    from beartype.typing import (
        Annotated,
        Callable,
        Collection,
        Container,
        Counter,
        Deque,
        Dict,
        FrozenSet,
        KeysView,
        List,
        Mapping,
        MutableMapping,
        MutableSequence,
        Sequence,
        Set,
        Sized,
        Tuple,
        Type,
        Union,
        ValuesView,
    )
    from beartype.vale import IsInstance
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_11
    from beartype_test.a00_unit.data.data_type import (
        Class,
        ClassCollection,
        ClassContainer,
        ClassDict,
        ClassList,
        ClassMapping,
        ClassMappingSequence,
        ClassMutableMapping,
        ClassMutableSequence,
        ClassSequence,
        ClassSized,
    )
    from beartype_test.a00_unit.data.func.data_func import (
        func_argless_return_unhinted,
        func_argless_return_hinted,
        func_args_unhinted_return_unhinted,
        func_args_flex_mandatory_kwonly_mandatory_hinted_return_hinted,
        func_args_flex_mandatory_optional_hinted_return_unhinted,
        func_args_flex_mandatory_varkw_hinted_return_hinted,
        func_args_flex_mandatory_varpos_hinted_return_hinted,
        func_args_unhinted_return_hinted,
    )
    from beartype_test._util.module.pytmodtyping import (
        import_typing_attr_or_none_safe)
    from collections import (
        Counter as CounterType,
        deque,
    )

    # ..................{ LOCALS ~ hint                      }..................
    # PEP 612-compliant "Concatenate" type hint factory and "ParamSpec" class
    # imported from either the standard "typing" or third-party
    # "typing_extensions" modules if importable from at least one of those
    # modules *OR* "None" otherwise.
    Concatenate = import_typing_attr_or_none_safe('Concatenate')
    ParamSpec = import_typing_attr_or_none_safe('ParamSpec')

    # ..................{ LOCALS ~ O(1)                      }..................
    # List of all type hint inference cases (i.e., 2-tuples "(obj, hint)"
    # describing the type hint matching an arbitrary object) specific to O(1)
    # constant-time type hint inference. Since this strategy achieves O(1)
    # constant time by inferring type hints pseudo-randomly, this list should
    # *ONLY* contain containers whose items are all of the same types.
    CASES_STRATEGY_O1 = [
        # ..................{ NON-PEP                        }..................
        # Builtin scalars are annotated as the builtin types of those scalars.
        ('Of his frail exultation shall be spent', str),
        (b'Mingling its solemn song, whilst the broad river', bytes),
        (True, bool),
        (73, int),
        (94.0251, float),

        # An instance of a PEP-noncompliant class (i.e., a class *NOT* covered
        # by an existing PEP standard) satisfying no broader type hint is
        # annotated simply as that class.
        (Class(), Class),

        # ..................{ PEP                            }..................
        # Any arbitrary type hint is annotated as itself.
        (Union[int, List[bytes]], Union[int, List[bytes]]),

        # ..................{ PEP 484                        }..................
        # The "None" singleton is annotated as itself under PEP 484.
        (None, None),

        # ..................{ PEP [484|585]                  }..................
        # A class is annotated as the PEP 484- or 585-compliant "type" supertype
        # subscripted by that class.
        (Class, Type[Class]),

        # ..................{ PEP [484|585] ~ abc            }..................
        # An instance of a PEP-noncompliant class (i.e., a class *NOT* covered
        # by an existing PEP standard) satisfying a standard "collections.abc"
        # protocol is annotated as the narrowest such protocol. Moreover, for
        # each instance that is a container (i.e., satisfies the standard
        # "collections.abc.Container" protocol), that annotation is subscripted
        # by a child type hint annotating the items of that container.

        # "collections.abc.MutableMapping" protocol hierarchy.
        (
            ClassMapping({
                'Branchless and blasted,': b'clenched with grasping roots',
                'The unwilling soil.': b'A gradual change was here,',
            }),
            # The narrowest "collections.abc" protocol matching this object
            # annotated to require an instance of this object's type.
            Annotated[Mapping[str, bytes], IsInstance[ClassMapping]],
        ),
        (
            ClassMutableMapping({
                b'Yet ghastly.': 'For, as fast years flow away,',
                b'The smooth brow gathers,': 'and the hair grows thin',
            }),
            # The narrowest "collections.abc" protocol matching this object
            # annotated to require an instance of this object's type.
            Annotated[
                MutableMapping[bytes, str], IsInstance[ClassMutableMapping]],
        ),

        # "collections.abc.MutableSequence" protocol hierarchy.
        (
            ClassContainer([
                b'Beneath the shade of trees', b'beside the flow']),
            # The narrowest "collections.abc" protocol matching this object
            # annotated to require an instance of this object's type.
            Annotated[Container, IsInstance[ClassContainer]],
        ),
        (
            ClassCollection([
                'Of the wild babbling rivulet;', 'and now']),
            # The narrowest "collections.abc" protocol matching this object
            # annotated to require an instance of this object's type.
            Annotated[Collection[str], IsInstance[ClassCollection]],
        ),
        (
            ClassSequence([
                b'He must descend.', b'With rapid steps he went']),
            # The narrowest "collections.abc" protocol matching this object
            # annotated to require an instance of this object's type.
            Annotated[Sequence[bytes], IsInstance[ClassSequence]],
        ),
        (
            ClassMutableSequence([
                'Forgetful of the grave', 'where', 'when the flame']),
            # The narrowest "collections.abc" protocol matching this object
            # annotated to require an instance of this object's type.
            Annotated[
                MutableSequence[str], IsInstance[ClassMutableSequence]],
        ),
        # Object satisfying both the "collections.abc.Mapping" *AND*
        # "collections.abc.Sequence" protocols, exercising various edge cases
        # with respect to protocol overlap and ambiguity.
        (
            ClassMappingSequence([
                b'And white,', b'and where irradiate dewy eyes',]),
            # The narrowest "collections.abc" protocol matching this object
            # annotated to require an instance of this object's type.
            Annotated[Sequence[bytes], IsInstance[ClassMappingSequence]],
        ),

        # "collections.abc.Sized" protocol.
        (
            ClassSized("The forest's solemn canopies were changed"),
            # The narrowest "collections.abc" protocol matching this object
            # annotated to require an instance of this object's type.
            Annotated[Sized, IsInstance[ClassSized]],
        ),

        # ..................{ PEP [484|585] ~ callable       }..................
        #FIXME: Additionally test:
        #* A pure-Python function isomorphically wrapping another function.
        #* A pure-Python function non-isomorphically wrapping another function.

        # A C-based callable is annotated as the unsubscripted PEP 484- or
        # 585-compliant "Callable" type.
        (
            iter,
            Callable,
        ),

        # A pure-Python callable annotated by *NO* return type hint accepting
        # *NO* parameters is annotated as the PEP 484- or 585-compliant
        # "Callable" type subscripted by the empty list followed by the
        # ignorable "object" superclass.
        (
            func_argless_return_unhinted,
            Callable[[], object],
        ),

        # A pure-Python callable annotated by a return type hint accepting *NO*
        # parameters is annotated as the PEP 484- or 585-compliant "Callable"
        # type subscripted by the empty list followed by that hint.
        (
            func_argless_return_hinted,
            Callable[[], str],
        ),

        # An unannotated pure-Python callable is annotated as the PEP 484- or
        # 585-compliant "Callable" type subscripted by either...
        (
            func_args_unhinted_return_unhinted,
            (
                # If the active Python interpreter targets Python >= 3.11 and
                # thus supports PEP 612:
                # * The PEP 612-compliant "typing(|_extensions).Concatenate"
                #   hint factory subscripted by one "object" child type hint for
                #   each mandatory positional-only or flexible parameter
                #   accepted by that callable followed by a trailing ellipsis.
                # * The ignorable "object" superclass, matching the return.
                Callable[Concatenate[object, object, ...], object]
                if IS_PYTHON_AT_LEAST_3_11 else
                # Else, the active Python interpreter targets Python < 3.11 and
                # thus fails to support PEP 612. In this case, the unsubscripted
                # PEP 484- or 585-compliant "Callable" type.
                Callable
            )
        ),

        # A pure-Python callable annotated by a return type hint and *NO*
        # parameter type hints is annotated as the PEP 484- or 585-compliant
        # "Callable" type subscripted by either...
        (
            func_args_unhinted_return_hinted,
            (
                # If the active Python interpreter targets Python >= 3.11 and
                # thus supports PEP 612:
                # * The PEP 612-compliant "typing(|_extensions).Concatenate"
                #   hint factory subscripted by one "object" child type hint for
                #   each mandatory positional-only or flexible parameter
                #   accepted by that callable followed by a trailing ellipsis.
                # * That return type hint.
                Callable[Concatenate[object, object, ...], str]
                if IS_PYTHON_AT_LEAST_3_11 else
                # Else, the active Python interpreter targets Python < 3.11 and
                # thus fails to support PEP 612. In this case, an ellipsis
                # followed by that return type hint.
                Callable[..., str]
            )
        ),

        # A pure-Python callable annotated by *NO* return type hint and one or
        # more mandatory and optional flexible parameter type hints is annotated
        # as the PEP 484- or 585-compliant "Callable" type subscripted by
        # either...
        (
            func_args_flex_mandatory_optional_hinted_return_unhinted,
            (
                # If the active Python interpreter targets Python >= 3.11 and
                # thus supports PEP 612:
                # * The PEP 612-compliant "typing(|_extensions).Concatenate"
                #   hint factory subscripted by these mandatory parameter type
                #   hints followed by a trailing ellipsis.
                # * The ignorable "object" superclass, matching the return.
                Callable[Concatenate[int, bytes, ...], object]
                if IS_PYTHON_AT_LEAST_3_11 else
                # Else, the active Python interpreter targets Python < 3.11 and
                # thus fails to support PEP 612. In this case, the unsubscripted
                # PEP 484- or 585-compliant "Callable" type.
                Callable
            )
        ),

        # A pure-Python callable annotated by a return type hint and one or
        # more mandatory flexible parameter type hints and a variadic positional
        # parameter type hint is annotated as the PEP 484- or 585-compliant
        # "Callable" type subscripted by either...
        (
            func_args_flex_mandatory_varpos_hinted_return_hinted,
            (
                # If the active Python interpreter targets Python >= 3.11 and
                # thus supports PEP 612:
                # * The PEP 612-compliant "typing(|_extensions).Concatenate"
                #   hint factory subscripted by these mandatory parameter type
                #   hints followed by a trailing ellipsis.
                # * That return type hint.
                Callable[Concatenate[bool, float, ...], str]
                if IS_PYTHON_AT_LEAST_3_11 else
                # Else, the active Python interpreter targets Python < 3.11 and
                # thus fails to support PEP 612. In this case, the PEP 484- or
                # 585-compliant "Callable" type subscripted by an ellipsis and
                # that return type hint.
                Callable[..., str]
            )
        ),

        # A pure-Python callable annotated by a return type hint and one or
        # more mandatory flexible parameter type hints and a variadic keyword
        # parameter type hint is annotated as the PEP 484- or 585-compliant
        # "Callable" type subscripted by either...
        (
            func_args_flex_mandatory_varkw_hinted_return_hinted,
            (
                # If the active Python interpreter targets Python >= 3.11 and
                # thus supports PEP 612:
                # * The PEP 612-compliant "typing(|_extensions).Concatenate"
                #   hint factory subscripted by these mandatory parameter type
                #   hints followed by a trailing ellipsis.
                # * That return type hint.
                Callable[Concatenate[str, int, ...], bytes]
                if IS_PYTHON_AT_LEAST_3_11 else
                # Else, the active Python interpreter targets Python < 3.11 and
                # thus fails to support PEP 612. In this case, the PEP 484- or
                # 585-compliant "Callable" type subscripted by an ellipsis and
                # that return type hint.
                Callable[..., bytes]
            )
        ),

        # A pure-Python callable annotated by a return type hint and one or
        # more mandatory flexible parameter type hints and a mandatory
        # keyword-only parameter type hint is annotated as the PEP 484- or
        # 585-compliant "Callable" type subscripted by either...
        (
            func_args_flex_mandatory_kwonly_mandatory_hinted_return_hinted,
            (
                # If the active Python interpreter targets Python >= 3.11 and
                # thus supports PEP 612:
                # * The PEP 612-compliant "typing(|_extensions).Concatenate"
                #   hint factory subscripted by these mandatory parameter type
                #   hints followed by a trailing ellipsis.
                # * That return type hint.
                Callable[Concatenate[float, int, ...], bytes]
                if IS_PYTHON_AT_LEAST_3_11 else
                # Else, the active Python interpreter targets Python < 3.11 and
                # thus fails to support PEP 612. In this case, the PEP 484- or
                # 585-compliant "Callable" type subscripted by an ellipsis and
                # that return type hint.
                Callable[..., bytes]
            )
        ),

        # ..................{ PEP [484|585] ~ counter        }..................
        # A counter of items all of the same class is annotated as the PEP 484-
        # or 585-compliant "Counter" type subscripted by that class.
        (
            CounterType({
                'For the uniform and': 27, 'lightsome evening sky.': 89,}),
            Counter[str],
        ),

        # ..................{ PEP [484|585] ~ deque          }..................
        # A deque of items all of the same class is annotated as the PEP 484- or
        # 585-compliant "deque" type subscripted by that class.
        (
            deque(('On the green moss his tremulous step,', 'that caught',)),
            Deque[str],
        ),

        # ..................{ PEP [484|585] ~ dict           }..................
        #FIXME: Validate that all other empty data structures behave as
        #expected, please. *sigh*

        # The empty dictionary is annotated as the unsubscripted PEP 484- or
        # 585-compliant "dict" type.
        (
            {},
            Dict,
        ),

        # A dictionary of one item is annotated as the PEP 484- or 585-compliant
        # "dict" type subscripted by the classes of that item.
        (
            {'In wanton sport, those bright leaves,': b'whose decay,',},
            Dict[str, bytes],
        ),

        # A dictionary of two or more items all of the same class is annotated
        # as the PEP 484- or 585-compliant "dict" type subscripted by that
        # class.
        (
            {b'Grey rocks': 'did peep from', b'the spare moss,': 'and stemmed'},
            Dict[bytes, str],
        ),

        # ..................{ PEP [484|585] ~ frozenset      }..................
        # A frozen set of items all of the same class is annotated as the PEP
        # 484- or 585-compliant "frozenset" type subscripted by that class.
        (
            frozenset({"I' the passing wind!", 'Beside the grassy shore',}),
            FrozenSet[str],
        ),

        # ..................{ PEP [484|585] ~ keysview       }..................
        # A keys view of keys all of the same class is annotated as the PEP 484-
        # or 585-compliant "KeysView" type subscripted by that class.
        ({
            b'What oozy cavern': 'or what wandering cloud expose',
            b'Contains thy waters,': 'as the universe',
        }.keys(), KeysView[bytes]),

        # ..................{ PEP [484|585] ~ list           }..................
        # An empty list is annotated as the unsubscripted PEP 484- or
        # 585-compliant "list" type hint factory.
        ([], List),

        # A list of items all of the same type is annotated as the PEP 484- or
        # 585-compliant "list" type subscripted by that type.
        (['expose', 'extreme', 'explosions!',], List[str]),

        # ..................{ PEP [484|585] ~ set            }..................
        # A set of items all of the same class is annotated as the PEP 484- or
        # 585-compliant "set" type subscripted by that class.
        ({'Of the small stream he went;', 'he did impress',}, Set[str]),

        # ..................{ PEP [484|585] ~ tuple          }..................
        # A subjectively small root tuple (i.e., tuple *NOT* contained in a
        # larger container and containing a fairly small number of items) is
        # annotated as the fixed-length PEP 484- and 585-compliant "tuple" type
        # iteratively subscripted by type hints matching the types of all items.
        (
            ('Foaming and hurrying o', b'er its rugged path', 94, 4.3618,),
            Tuple[str, bytes, int, float],
        ),

        # A subjectively large root tuple (i.e., tuple *NOT* contained in a
        # larger container but containing a fairly large number of items all of
        # the same type) is annotated as the variadic PEP 484- and
        # 585-compliant "tuple" type subscripted by that type.
        (
            (
                'The wilds to love tranquillity. One step,'
                'One human step alone, has ever broken',
                'The stillness of its solitude:—one voice',
                'Alone inspired its echoes;—even that voice',
            ) * 4,
            Tuple[str, ...]
        ),

        # ..................{ PEP [484|585] ~ valuesview     }..................
        # A values view of values all of the same class is annotated as the PEP
        # 484- or 585-compliant "ValuesView" type subscripted by that class.
        ({
            'Tell where these living thoughts reside,': b'when stretched',
            'Upon thy flowers': b'my bloodless limbs shall waste',
        }.values(), ValuesView[bytes]),

        # ..................{ PEP 585 ~ builtin              }..................
        # An instance of a user-defined class subclassing a PEP 585-compliant
        # builtin collection type (e.g., "dict", "list") is annotated as that
        # user-defined class subscripted by a child type hint annotating the
        # items of that collection.

        # A user-defined dictionary of keys all of the same class and values all
        # of another class is annotated as the PEP 585-compliant "dict" type
        # sequentially subscripted by those classes.
        (
            ClassDict({
                b'Threw their thin shadows': 'down the rugged slope',
                b'And nought but': 'gnarlèd roots of ancient pines',
            }),
            # This user-defined dictionary subscripted by these child key and
            # value type hints.
            ClassDict[bytes, str],
        ),

        # A user-defined list of items all of the same class is annotated as the
        # PEP 585-compliant "list" type subscripted by that class.
        (
            ClassList((
                b'The struggling brook:', b'tall spires of windlestrae',)),
            # This user-defined list subscripted by this child type hint.
            ClassList[bytes],
        ),
    ]

    # ..................{ VERSIONS                           }..................
    # If the PEP 612-compliant "ParamSpec" class is importable from either the
    # standard "typing" or third-party "typing_extensions" modules...
    if ParamSpec is not None:
        # Defer attribute-specific imports.
        from beartype_test.a00_unit.data.func.data_pep612 import (
            P,
            func_args_flex_mandatory_kwonly_optional_paramspec_return_hinted,
            func_args_flex_mandatory_paramspec_distinct_return_hinted,
            func_args_flex_mandatory_paramspec_return_hinted,
            func_args_flex_mandatory_paramspec_varkwonly_return_hinted,
            func_args_flex_mandatory_paramspec_varposonly_return_hinted,
            func_args_paramspec_return_hinted,
        )

        # Extend this list with "ParamSpec"-specific cases.
        CASES_STRATEGY_O1.extend((
            # ..................{ PEP 612                    }..................
            # A pure-Python callable annotated by a return type hint accepting
            # variadic positional and keyword parameters annotated as the
            # corresponding instance variables of a PEP 612-compliant parameter
            # specification is annotated as the PEP 484- or 585-compliant
            # "Callable" type subscripted by that parameter specification
            # followed by that return type hint.
            (
                func_args_paramspec_return_hinted,
                Callable[P, bytes],
            ),

            # A pure-Python callable annotated by a return type hint accepting
            # one or more mandatory flexible parameter type hints *AND* variadic
            # positional and keyword parameters annotated as the corresponding
            # instance variables of a PEP 612-compliant parameter specification
            # is annotated as the PEP 484- or 585-compliant "Callable" type
            # subscripted by:
            # * The PEP 612-compliant "typing(|_extensions).Concatenate" hint
            #   factory subscripted by these mandatory parameter type hints
            #   followed by that parameter specification.
            # * That return type hint.
            (
                func_args_flex_mandatory_paramspec_return_hinted,
                Callable[Concatenate[int, bool, P], str],
            ),

            # A pure-Python callable annotated by a return type hint accepting
            # one or more mandatory flexible parameter type hints and a variadic
            # positional parameter annotated as the corresponding instance
            # variables of a PEP 612-compliant parameter specification with *NO*
            # variadic keyword parameter is annotated as either...
            (
                func_args_flex_mandatory_paramspec_varposonly_return_hinted,
                (
                    # If the active Python interpreter targets Python >= 3.11
                    # and thus supports PEP 612, the PEP 484- or 585-compliant
                    # "Callable" type subscripted by:
                    # * The PEP 612-compliant "typing(|_extensions).Concatenate"
                    #   hint factory subscripted by these mandatory parameter
                    #   type hints followed by an ellipsis.
                    # * That return type hint.
                    Callable[Concatenate[int, float, ...], bytes]
                    if IS_PYTHON_AT_LEAST_3_11 else
                    # Else, the active Python interpreter targets Python < 3.11
                    # and thus fails to support PEP 612. In this case, the PEP
                    # 484- or 585-compliant "Callable" type subscripted by an
                    # ellipsis and that return type hint.
                    Callable[..., bytes]
                ),
            ),

            # A pure-Python callable annotated by a return type hint accepting
            # one or more mandatory flexible parameter type hints and a variadic
            # keyword parameter annotated as the corresponding instance
            # variables of a PEP 612-compliant parameter specification with *NO*
            # variadic positional parameter is annotated as either...
            (
                func_args_flex_mandatory_paramspec_varkwonly_return_hinted,
                (
                    # If the active Python interpreter targets Python >= 3.11
                    # and thus supports PEP 612, the PEP 484- or 585-compliant
                    # "Callable" type subscripted by:
                    # * The PEP 612-compliant "typing(|_extensions).Concatenate"
                    #   hint factory subscripted by these mandatory parameter
                    #   type hints followed by an ellipsis.
                    # * That return type hint.
                    Callable[Concatenate[bool, bytes, ...], str]
                    if IS_PYTHON_AT_LEAST_3_11 else
                    # Else, the active Python interpreter targets Python < 3.11
                    # and thus fails to support PEP 612. In this case, the PEP
                    # 484- or 585-compliant "Callable" type subscripted by an
                    # ellipsis and that return type hint.
                    Callable[..., str]
                ),
            ),

            # A pure-Python callable annotated by a return type hint accepting
            # one or more mandatory flexible parameter type hints, a variadic
            # positional parameter annotated as the corresponding instance
            # variables of a PEP 612-compliant parameter specification, and a
            # variadic keyword parameter annotated as the corresponding instance
            # variables of a different parameter specification is annotated as
            # either...
            (
                func_args_flex_mandatory_paramspec_distinct_return_hinted,
                (
                    # If the active Python interpreter targets Python >= 3.11
                    # and thus supports PEP 612, the PEP 484- or 585-compliant
                    # "Callable" type subscripted by:
                    # * The PEP 612-compliant "typing(|_extensions).Concatenate"
                    #   hint factory subscripted by these mandatory parameter
                    #   type hints followed by an ellipsis.
                    # * That return type hint.
                    Callable[Concatenate[float, int, ...], bytes]
                    if IS_PYTHON_AT_LEAST_3_11 else
                    # Else, the active Python interpreter targets Python < 3.11
                    # and thus fails to support PEP 612. In this case, the PEP
                    # 484- or 585-compliant "Callable" type subscripted by an
                    # ellipsis and that return type hint.
                    Callable[..., bytes]
                ),
            ),

            # A pure-Python callable annotated by a return type hint accepting
            # one or more mandatory flexible parameter type hints, an optional
            # flexible parameter type hint, *AND* variadic positional and
            # keyword parameters annotated as the corresponding instance
            # variables of a PEP 612-compliant parameter specification is
            # annotated as either...
            (
                func_args_flex_mandatory_kwonly_optional_paramspec_return_hinted,
                (
                    # If the active Python interpreter targets Python >= 3.11
                    # and thus supports PEP 612, the PEP 484- or 585-compliant
                    # "Callable" type subscripted by:
                    # * The PEP 612-compliant "typing(|_extensions).Concatenate"
                    #   hint factory subscripted by these mandatory parameter
                    #   type hints followed by an ellipsis.
                    # * That return type hint.
                    Callable[Concatenate[float, complex, ...], bytes]
                    if IS_PYTHON_AT_LEAST_3_11 else
                    # Else, the active Python interpreter targets Python < 3.11
                    # and thus fails to support PEP 612. In this case, the PEP
                    # 484- or 585-compliant "Callable" type subscripted by an
                    # ellipsis and that return type hint.
                    Callable[..., bytes]
                ),
            ),
        ))
    # Else, the PEP 612-compliant "ParamSpec" class is unimportable.

    # ..................{ LOCALS ~ O(N)                      }..................
    # List of all type hint inference cases (i.e., 2-tuples "(obj, hint)"
    # describing the type hint matching an arbitrary object) specific to O(n)
    # linear-time type hint inference. Since this strategy deterministically
    # infers type hints in a fully recursive manner, this list may contain
    # containers whose items are of differing types.
    CASES_STRATEGY_ON = CASES_STRATEGY_O1 + [
        # ..................{ PEP [484|585] ~ list           }..................
        # A list of items all of differing types is annotated as the PEP 484-
        # and 585-compliant "list" type subscripted by a PEP 604-compliant union
        # hint of those types -- including...
        ([
            # A string.
            'Thy dazzling waves, thy loud and hollow gulfs,',
            # A byte string.
            b'Thy searchless fountain, and invisible course',
            # An integer.
            len('Have each their type in me: and the wide sky,'),
            # A list of strings.
            ['And measureless ocean', 'may declare', 'as soon',],
        ], list[str | bytes | int | list[str]]),

        # ..................{ PEP [484|585] ~ tuple          }..................
        # A subjectively large root tuple (i.e., tuple *NOT* contained in a
        # larger container but containing a fairly large number of items all of
        # differing types) is annotated as the variadic PEP 484- and
        # 585-compliant "tuple" type subscripted by a PEP 604-compliant union
        # hint of those types.
        (
            (
                'Fell into that immeasurable void,',
                b'Scattering its waters to the passing winds.',
                37,
                9.1725,
            ) * 4,
            tuple[str | bytes | int | float, ...],
        ),
    ]

    # ..................{ DICTS ~ cases                      }..................
    # Dictionary mapping from each relevant beartype configuration to a
    # corresponding list of all type hint inference cases specific to that
    # configuration (i.e., 2-tuples "(obj, hint)" describing the type hint
    # matching an arbitrary object) to be returned by this fixture.
    CONF_TO_CASES = {
        # O(1) constant-time type hint inference.
        BeartypeConf(strategy=BeartypeStrategy.O1): CASES_STRATEGY_O1,

        # O(n) linear-time type hint inference.
        BeartypeConf(strategy=BeartypeStrategy.On): CASES_STRATEGY_ON,
    }

    # ..................{ RETURN                             }..................
    # Return this dictionary.
    return CONF_TO_CASES



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **shell** (i.e., external low-level command-line environment
encapsulating the active Python interpreter as a parent process of this
interpreter) utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from os import environ

# ....................{ TESTERS                            }....................
get_shell_var_value_or_none = environ.get
'''
String value of the shell environment variable with the passed name if the
parent shell defines this variable *or* :data:`None` otherwise (i.e., if the
parent shell does *not* define this variable).

Caveats
-------
**This getter is a human-readable alias of the comparable**
:func:`os.getenv` **function and** :meth:`os.environ.get` **method.** This
getter exists only for disambiguity and clarity. This getter is *not* an alias
of the :meth:`os.environ.__getitem__` dunder method, which raises a
:exc:`KeyError` exception rather than returns :data:`None` if the parent shell
fails to define this variable.

See Also
--------
https://stackoverflow.com/a/41626355/2809027
    StackOverflow answer strongly inspiring this alias.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484` and :pep:`585` **type hint utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep484585.pep484585` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ getter                     }....................
def test_get_hint_pep484585_args() -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep484585.pep484585.get_hint_pep484585_args`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep585Exception
    from beartype.typing import (
        Dict,
        List,
    )
    from beartype._util.hint.pep.proposal.pep484585.pep484585 import (
        get_hint_pep484585_args)
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert this getter returns the expected child type hint for a PEP 484-
    # and 585-compliant type hint properly subscripted by one child type hint.
    assert get_hint_pep484585_args(
        hint=List[bool], args_len=1, exception_prefix='') == bool

    # Assert this getter returns the expected child type hints for a PEP 484-
    # and 585-compliant type hint properly subscripted by two or more child type
    # hints.
    assert get_hint_pep484585_args(
        hint=Dict[str, int], args_len=2, exception_prefix='') == (str, int)

    # ....................{ FAIL                           }....................
    # Assert this tester raises the expected exception for an arbitrary type
    # hint passed invalid argument lengths.
    with raises(AssertionError):
        get_hint_pep484585_args(
            hint='He fled. Red morning dawned upon his flight,',
            args_len=0.01,
            exception_prefix='',
        )
    with raises(AssertionError):
        get_hint_pep484585_args(
            hint='Shedding the mockery of its vital hues',
            args_len=0,
            exception_prefix='',
        )

    # Assert this tester raises the expected exception for a PEP
    # 585-compliant dictionary type hint improperly subscripted by an
    # unexpected number of child type hints.
    with raises(BeartypeDecorHintPep585Exception):
        get_hint_pep484585_args(
            hint=Dict[str], args_len=2, exception_prefix='')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`749` **type hint utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep749` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ getter                     }....................
def test_get_hint_pep749_subhint() -> None:
    '''
    Test the pair of private
    :mod:`beartype._util.hint.pep.proposal.pep749.get_hint_pep749_subhint_mandatory`
    and
    :mod:`beartype._util.hint.pep.proposal.pep749.get_hint_pep749_subhint_optional`
    getters.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep749Exception
    from beartype.typing import Optional
    from beartype._data.kind.datakindiota import SENTINEL
    from beartype._data.typing.datatypingport import Hint
    from beartype._util.hint.pep.proposal.pep749 import (
        get_hint_pep749_subhint_mandatory,
        get_hint_pep749_subhint_optional,
    )
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_14
    from pytest import raises

    # ....................{ CONSTANTS                      }....................
    # Arbitrary object to be used as the null subhint value (i.e., object
    # signifying a subhint to be unspecified) below.
    SUBHINT_VALUE_NULL = 0xFEEDFACE

    # ....................{ DESCRIPTORS                    }....................
    class DescriptorNameErrorRaiser(object):
        '''
        Arbitrary descriptor whose :meth:`__get__` method unconditionally raises
        a builtin :exc:`NameError` exception, simulating a subhint whose static
        value contains one or more unquoted forward references.
        '''

        def __get__(self, obj: object, objtype: Optional[type] = None):
            raise NameError('Oft made Hyperion ache. His palace bright')

    # ....................{ CLASSES                        }....................
    class UponTheFirstToll(object):
        '''
        Arbitrary class defining arbitrary static and dynamic subhints to be
        retrieved by this getter.
        '''

        # ....................{ STATIC                     }....................
        of_his_passing_bell = 'Upon the first toll of his passing bell,'
        '''
        Arbitrary class variable simulating a subhint whose static value
        contains *no* unquoted forward references and is *not* the null.
        '''


        oft_made_hyperion_ache = SUBHINT_VALUE_NULL
        '''
        Arbitrary class variable simulating a subhint whose static value
        contains *no* unquoted forward references but is the null.
        '''

        # ....................{ DYNAMIC                    }....................
        portioned_to_a_giant_nerve = DescriptorNameErrorRaiser()
        '''
        Arbitrary class variable simulating a subhint whose static value
        contains *no* unquoted forward references.
        '''


        def or_prophesyings(self, hint_format: int) -> Hint:
            '''
            Arbitrary method simulating a subhint whose static value contains
            one or more unquoted forward references and thus requires an
            :pep:`749`-compliant evaluator method.
            '''

            return 'Or prophesyings of the midnight lamp;'

    # ....................{ LOCALS                         }....................
    # Arbitrary instance of this class.
    but_horrors = UponTheFirstToll()

    # ....................{ PASS                           }....................
    # Assert these getters when passed a hint returns the subhint with the
    # passed static attribute name (when this subhint contains *NO* unquoted
    # forward references and is *NOT* the null).
    assert get_hint_pep749_subhint_mandatory(
        hint=but_horrors,
        subhint_name_dynamic='or_prophesyings',
        subhint_name_static='of_his_passing_bell',
    ) is get_hint_pep749_subhint_optional(
        hint=but_horrors,
        subhint_name_dynamic='or_prophesyings',
        subhint_name_static='of_his_passing_bell',
        subhint_value_null=SUBHINT_VALUE_NULL,
    ) is UponTheFirstToll.of_his_passing_bell

    # Assert this getter when passed a hint returns the sentinel placeholder
    # (when this subhint contains *NO* unquoted forward references but is the
    # null).
    assert get_hint_pep749_subhint_optional(
        hint=but_horrors,
        subhint_name_dynamic='or_prophesyings',
        subhint_name_static='oft_made_hyperion_ache',
        subhint_value_null=SUBHINT_VALUE_NULL,
    ) is SENTINEL

    # ....................{ FAIL                           }....................
    # Assert this getter raises the expected exception when passed a hint that
    # defines *NO* subhint with the passed static or dynamic attribute names.
    with raises(BeartypeDecorHintPep749Exception):
        get_hint_pep749_subhint_mandatory(
            hint="But horrors, portion'd to a giant nerve,",
            subhint_name_dynamic='or_prophesyings',
            subhint_name_static='of_his_passing_bell',
        )

    # ....................{ VERSIONS                       }....................
    # If the active Python interpreter targets Python >= 3.14 and thus supports
    # PEP 749...
    if IS_PYTHON_AT_LEAST_3_14:
        # Defer version-specific imports.
        from annotationlib import Format

        # Assert these getters when passed a hint returns the subhint with the
        # passed dynamic attribute name (when this subhint contains one or more
        # unquoted forward references and is *NOT* the null).
        assert get_hint_pep749_subhint_mandatory(
            hint=but_horrors,
            subhint_name_dynamic='or_prophesyings',
            subhint_name_static='portioned_to_a_giant_nerve',
        ) is get_hint_pep749_subhint_optional(
            hint=but_horrors,
            subhint_name_dynamic='or_prophesyings',
            subhint_name_static='portioned_to_a_giant_nerve',
            subhint_value_null=SUBHINT_VALUE_NULL,
        ) == 'Or prophesyings of the midnight lamp;'

        # Assert these getters when passed a hint returns the stringified
        # subhint with the passed dynamic attribute name (when this subhint
        # contains one or more unquoted forward references and is *NOT* the
        # null) under the string format.
        assert get_hint_pep749_subhint_optional(
            hint=but_horrors,
            subhint_name_dynamic='or_prophesyings',
            subhint_name_static='portioned_to_a_giant_nerve',
            subhint_value_null=SUBHINT_VALUE_NULL,
            hint_format=Format.STRING,
        ) == 'Or prophesyings of the midnight lamp;'

        # Assert this getter raises the expected exception when passed a hint
        # whose subhint contains one or more unquoted forward references under
        # the value format.
        with raises(NameError):
            assert get_hint_pep749_subhint_optional(
                hint=but_horrors,
                subhint_name_dynamic='or_prophesyings',
                subhint_name_static='portioned_to_a_giant_nerve',
                subhint_value_null=SUBHINT_VALUE_NULL,
                hint_format=Format.VALUE,
            )
    # Else, this interpreter targets Python < 3.14 and thus fails to support PEP
    # 749.



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`612`-compliant **parameter specification** (i.e.,
:obj:`typing.ParamSpec` type hints and instance variables of those hints)
utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep612Exception
from beartype.typing import (
    Callable,
    Optional,
    Union,
)
from beartype._cave._cavefast import (
    EllipsisType,
    HintPep612ParamSpecType,
    HintPep612ParamSpecArgType,
    HintPep612ParamSpecKwargType,
)
from beartype._check.metadata.metadecor import BeartypeDecorMeta
from beartype._data.typing.datatypingport import (
    Hint,
    ListHints,
)
from beartype._data.hint.sign.datahintsigncls import HintSign
from beartype._data.hint.sign.datahintsigns import (
    HintSignParamSpecArgs,
    HintSignParamSpecKwargs,
)
from beartype._util.api.standard.utiltyping import import_typing_attr_or_none
from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none
from beartype._util.func.arg.utilfuncargget import (
    get_func_arg_meta_variadic_keyword_or_none,
    get_func_arg_meta_variadic_positional_or_none,
)
from beartype._util.func.arg.utilfuncargiter import (
    ARG_META_INDEX_NAME,
    ArgKind,
)
from beartype._util.func.arg.utilfuncargtest import (
    is_func_arg_name_variadic_keyword,
    is_func_arg_name_variadic_positional,
)
from beartype._data.kind.datakindiota import SENTINEL

# ....................{ GETTERS                            }....................
def get_hint_pep612_paramspec(
    paramspec_var: Union[
        HintPep612ParamSpecArgType,
        HintPep612ParamSpecKwargType,
    ]
) -> HintPep612ParamSpecType:
    '''
    :pep:`612`-compliant **parameter specification** (i.e., low-level C-based
    :obj:`typing.ParamSpec` object) containing the passed **parameter
    specification variadic parameter instance variable** (i.e., low-level
    C-based :obj:`typing.ParamSpecArgs` object annotating a variadic parameter
    with syntax resembling ``*args: P.args`` and ``**kwargs: P.kwargs`` for
    ``P`` a low-level C-based :obj:`typing.ParamSpec` parent object) as a child
    instance variable.

    Parameters
    ----------
    paramspec_var : HintPep612ParamSpecVarTypes
        Parameter specification variadic parameter instance variable to be
        inspected.

    Returns
    -------
    typing.ParamSpec
        Parameter specification containing this variable.
    '''
    #FIXME: *lol*. This assert fails when using "typing_extensions.ParamSpec"
    #under Python < 3.10. Whatevahs! Just ignore quality assurance for now. \o/
    # assert isinstance(paramspec_var, HintPep612ParamSpecVarTypes), (
    #     f'{repr(paramspec_var)} not '
    #     f'PEP 612-compliant parameter specification variable.'
    # )

    # One-liners are fine. One-liners never cross the line! ¯\_(ツ)_/¯
    return paramspec_var.__origin__  # type: ignore[union-attr]

# ....................{ FACTORIES                          }....................
#FIXME: This should probably be memoized by @callable_cached. Probably. *sigh*
def make_hint_pep612_concatenate_list_or_none(
    hints_child_first: ListHints,
    hint_child_last: Union[Hint, EllipsisType],  # type: ignore[valid-type]
) -> Hint:
    '''
    :pep:`612`-compliant **parameter concatenation** (i.e., high-level
    pure-Python object created and returned by subscripting the standard
    :func:`typing.Concatenate` type hint factory) dynamically subscripted first
    by all child type hints in the passed ``hints_child_first`` list and then by
    the child type hint ``hint_child_last` if this hint factory is importable
    under the active Python interpreter *or* :data:`None` otherwise (i.e., if
    this hint factory is unimportable under the active Python interpreter).

    Parameters
    ----------
    hints_child_first : ListHints
        List of all leading child type hints to subscript the returned
        ``typing.Concatenate[...]`` type hint with.
    hint_child_last : Union[Hint, EllipsisType]
        Trailing child type hint to subscript the returned
        ``typing.Concatenate[...]`` type hint with. Note that both :pep:`612`
        *and* the runtime implementation of the :func:`typing.Concatenate` type
        hint factory require that this hint be either:

        * A **parameter specification** (i.e., low-level C-based
          :pep:`612`-compliant :obj:`typing.ParamSpec` object).
        * An **ellipsis** (i.e., ``...``), silently ignoring *all* remaining
          arbitrary trailing parameters accepted by the callable matched by the
          parent :pep:`484`- or :pep:`585`-compliant ``Callable[...]`` type
          hint.

    Returns
    -------
    Hint
        Either:

        * If :func:`typing.Concatenate` is importable, the
          ``typing.Concatenate[...]`` type hint subscripted by these child type
          hints.
        * Else, :data:`None`.

    Raises
    ------
    TypeError
        If ``hint_child_last`` is neither:

        * A parameter specification.
        * An ellipsis.
    '''
    assert isinstance(hints_child_first, list), (
        f'{repr(hints_child_first)} not list.')

    # "Concatenate[...]" hint to be returned, subscripted by these child hints.
    hint = None

    # PEP 612-compliant "typing(|_extensions).Concatenate" hint factory if
    # importable *OR* "None" otherwise.
    Concatenate = import_typing_attr_or_none('Concatenate')

    #FIXME: When Python 3.10 support is dropped, this can and should be
    #reduced to this elegant one-liner employing list unpacking:
    #    hint = Concatenate[*hints_child_first, hint_child_last]

    # If this hint factory is importable...
    if Concatenate is not None:
        # Tuple of all child hints to subscript "Concatenate[...]" with.
        #
        # Note that the Concatenate.__getitem__() implementation *REQUIRES* that
        # the passed parameter be a tuple. Sanity is out the window, folks.
        hints_child = tuple(hints_child_first) + (hint_child_last,)
        # print(f'hints_child_first: {hints_child_first}; hint_child_last: {hint_child_last}')

        # "Concatenate[...]" hint dynamically subscripted by this tuple.
        hint = Concatenate.__getitem__(hints_child)  # type: ignore[misc]
    # Else, this hint factory is unimportable. In this case, fallback to
    # returning "None".

    # Return this "Concatenate[...]" hint.
    return hint  # pyright: ignore

# ....................{ REDUCERS                           }....................
def reduce_hint_pep612_args(hint: object, **kwargs,) -> object:
    '''
    Reduce the passed :pep:`612`-compliant **parameter specification variadic
    positional parameter instance variable** (i.e., low-level C-based
    :obj:`typing.ParamSpecArgs` object annotating a variadic positional parameter
    with syntax resembling ``*args: P.args`` for ``P`` a low-level C-based
    :obj:`typing.ParamSpec` parent object) to the ignorable :class:`object`
    superclass.

    This reducer effectively ignores *all* :obj:`typing.ParamSpecArgs` objects.
    Although these objects do convey meaningful metadata and semantics, they do
    so in a manner uniquely suited to pure static type-checkers. Why? Because
    type-checking these objects requires tracking parameter specifications
    across disparate callable signatures separated - in the worst case - in both
    space and time. Although :mod:`beartype` could type-check these objects at a
    later date, doing so is highly non-trivial and thus deferred to the
    sacrificial blood sacrifice that is reading this. Sorry. That's you.

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : object
        :pep:`612`-compliant parameter specification variadic positional
        parameter type hint to be reduced.

    All remaining keyword parameters are passed as is to the lower-level
    :func:`._reduce_hint_pep612_args_or_kwargs` reducer.

    Returns
    -------
    object
        Lower-level type hint currently supported by :mod:`beartype`.

    Raises
    ------
    BeartypeDecorHintPep612Exception
        If this hint does *not* annotate a variadic positional parameter.

    See Also
    --------
    :func:`._reduce_hint_pep612_args_or_kwargs`
        Further details.
    '''

    # Defer to this lower-level general-purpose reducer.
    return _reduce_hint_pep612_args_or_kwargs(
        hint=hint,
        arg_kind_expected=ArgKind.VARIADIC_POSITIONAL,
        pith_name_label=_VAR_POS_PITH_NAME_LABEL,
        pith_name_syntax=_VAR_POS_PITH_NAME_SYNTAX,
        other_hint=get_hint_pep612_paramspec(hint).kwargs,  # type: ignore[attr-defined,arg-type]
        other_hint_sign=HintSignParamSpecKwargs,
        other_pith_name_getter=get_func_arg_meta_variadic_keyword_or_none,
        other_pith_name_label=_VAR_KW_PITH_NAME_LABEL,
        **kwargs
    )


def reduce_hint_pep612_kwargs(hint: object, **kwargs) -> object:
    '''
    Reduce the passed :pep:`612`-compliant **parameter specification variadic
    keyword parameter instance variable** (i.e., low-level C-based
    :obj:`typing.ParamSpecKwargs` object annotating a variadic keyword
    parameter with syntax resembling ``**kwargs: P.kwargs`` for ``P`` a low-level
    C-based :obj:`typing.ParamSpec` parent object) to the ignorable
    :class:`object` superclass.

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    See Also
    --------
    :func:`.reduce_hint_pep612_args`
        Further details.
    '''
    # print('!!HERE!!')

    # Defer to this lower-level general-purpose reducer.
    return _reduce_hint_pep612_args_or_kwargs(
        hint=hint,
        arg_kind_expected=ArgKind.VARIADIC_KEYWORD,
        pith_name_label=_VAR_KW_PITH_NAME_LABEL,
        pith_name_syntax=_VAR_KW_PITH_NAME_SYNTAX,
        other_hint=get_hint_pep612_paramspec(hint).args,  # type: ignore[attr-defined,arg-type]
        other_hint_sign=HintSignParamSpecArgs,
        other_pith_name_getter=get_func_arg_meta_variadic_positional_or_none,
        other_pith_name_label=_VAR_POS_PITH_NAME_LABEL,
        **kwargs
    )

# ....................{ PRIVATE ~ constants : positional   }....................
_VAR_POS_PITH_NAME_LABEL = 'positional'
'''
Human-readable substring describing a :pep:`612`-compliant parameter
specification variadic positional parameter type hint.
'''


_VAR_POS_PITH_NAME_SYNTAX = '*args'
'''
Standard machine-readable syntax declaring a variadic positional parameter.
'''


_VAR_POS_PITH_NAME_TESTER = is_func_arg_name_variadic_positional
'''
Human-readable substring validating that a :pep:`612`-compliant parameter
specification variadic positional parameter type hint annotates a variadic
positional parameter.
'''

# ....................{ PRIVATE ~ constants : keyword      }....................
_VAR_KW_PITH_NAME_LABEL = 'keyword'
'''
Human-readable substring describing a :pep:`612`-compliant parameter
specification variadic keyword parameter type hint.
'''


_VAR_KW_PITH_NAME_SYNTAX = '*args'
'''
Standard machine-readable syntax declaring a variadic positional parameter.
'''


_VAR_KW_PITH_NAME_TESTER = is_func_arg_name_variadic_keyword
'''
Human-readable substring validating that a :pep:`612`-compliant parameter
specification variadic keyword parameter type hint annotates a variadic keyword
parameter.
'''

# ....................{ PRIVATE ~ reducers                 }....................
def _reduce_hint_pep612_args_or_kwargs(
    # General-purpose parameters passed by the higher-level
    # beartype._check.convert._reduce.redmain.reduce_hint() reducer to this
    # lower-level reducer.
    hint: object,
    decor_meta: Optional[BeartypeDecorMeta],
    pith_name: Optional[str],
    arg_kind: Optional[ArgKind],
    exception_prefix: str,

    # PEP 612-specific parameters passed by the higher-level
    # reduce_hint_pep612_args() or reduce_hint_pep612_kwargs() reducer to this
    # lower-level reducer.
    arg_kind_expected: ArgKind,
    pith_name_label: str,
    pith_name_syntax: str,
    other_hint: object,
    other_hint_sign: HintSign,
    other_pith_name_getter: Callable,
    other_pith_name_label: str,

    # Ignorable general-purpose parameters passed by the higher-level
    # beartype._check.convert._reduce.redmain.reduce_hint() reducer *NOT* required
    # by this lower-level reducer.
    **kwargs
) -> object:
    '''
    Reduce the passed :pep:`612`-compliant **parameter specification variadic
    positional parameter instance variable** (i.e., low-level C-based
    :obj:`typing.ParamSpecArgs` object annotating a variadic positional parameter
    with syntax resembling ``*args: P.args`` for ``P`` a low-level C-based
    :obj:`typing.ParamSpec` parent object) to the ignorable :class:`object`
    superclass.

    This reducer effectively ignores *all* :obj:`typing.ParamSpecArgs` objects.
    Although these objects do convey meaningful metadata and semantics, they do
    so in a manner uniquely suited to pure static type-checkers. Why? Because
    type-checking these objects requires tracking parameter specifications
    across disparate callable signatures separated - in the worst case - in both
    space and time. Although :mod:`beartype` could type-check these objects at a
    later date, doing so is highly non-trivial and thus deferred to the
    sacrificial blood sacrifice that is reading this. Sorry. That's you.

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : object
        :pep:`612`-compliant parameter specification variadic positional or
        keyword parameter type hint to be reduced.
    decor_meta : Optional[BeartypeDecorMeta]
        Either:

        * If this hint annotates a parameter or return of some callable, the
          :mod:`beartype`-specific decorator metadata describing that callable.
        * Else, :data:`None`.
    pith_name : Optional[str]
        Either:

        * If this hint annotates a parameter of some callable, the name of that
          parameter.
        * If this hint annotates the return of some callable, ``"return"``.
        * Else, :data:`None`.

        Defaults to :data:`None`.
    arg_kind : Optional[ArgKind]
        Either:

        * If this hint annotates a parameter of some callable, that parameter's
          **kind** (i.e., :class:`.ArgKind` enumeration member conveying the
          syntactic class of that parameter, constraining how the callable
          declaring that parameter requires that parameter to be passed).
        * Else, :data:`None`.
    exception_prefix : str
        Human-readable substring prefixing exception messages raised by this
        reducer.
    arg_kind_expected : ArgKind
        **Expected parameter kind**, defined as either:

        * If this hint is expected to annotate a variadic positional parameter,
          the :attr:`.ArgKind.VARIADIC_POSITIONAL` enumeration member.
        * If this hint is expected to annotate a variadic keyword parameter,
          the :attr:`.ArgKind.VARIADIC_KEYWORD` enumeration member.
    pith_name_label : str
        Either:

        * If this hint annotates a variadic positional parameter, the
          human-readable substring ``"positional"``.
        * If this hint annotates a variadic keyword parameter, the
          human-readable substring ``"keyword"``.
    pith_name_syntax : str
        Either:

        * If this hint annotates a variadic positional parameter, the
          machine-readable substring ``"*args"``.
        * If this hint annotates a variadic keyword parameter, the
          machine-readable substring ``"**kwargs"``.
    other_hint : object
        Either:

        * If this hint annotates a variadic positional parameter, the
          corresponding hint annotating a variadic keyword parameter.
        * If this hint annotates a variadic keyword parameter, the
          corresponding hint annotating a variadic positional parameter.
    other_hint_sign : HintSign
        Either:

        * If this hint annotates a variadic positional parameter, the
          :data:`.HintSignParamSpecKwargs` sign.
        * If this hint annotates a variadic keyword parameter, the
          :data:`.HintSignParamSpecArgs` sign.
    other_pith_name_getter : Callable
        Either:

        * If this hint annotates a variadic positional parameter, the
          :func:`beartype._util.func.arg.utilfuncargget.get_func_arg_meta_variadic_keyword_or_none`
          getter.
        * If this hint annotates a variadic keyword parameter, the
          :func:`beartype._util.func.arg.utilfuncargget.get_func_arg_meta_variadic_positional_or_none`
          getter.
    other_pith_name_label : str
        Either:

        * If this hint annotates a variadic positional parameter, the
          human-readable substring ``"keyword"``.
        * If this hint annotates a variadic keyword parameter, the
          human-readable substring ``"positional"``.

    All remaining passed parameters are silently ignored.

    Returns
    -------
    object
        Lower-level type hint currently supported by :mod:`beartype`.

    Raises
    ------
    BeartypeDecorHintPep612Exception
        If this hint does *not* annotate a variadic positional parameter.
    '''
    assert isinstance(arg_kind_expected, ArgKind), (
        f'{repr(arg_kind_expected)} not argument kind.')
    assert isinstance(pith_name_label, str), (
        f'{repr(pith_name_label)} not string.')
    assert isinstance(pith_name_syntax, str), (
        f'{repr(pith_name_syntax)} not string.')
    assert isinstance(other_hint_sign, HintSign), (
        f'{repr(other_hint_sign)} not type hint sign.')
    assert callable(other_pith_name_getter), (
        f'{repr(other_pith_name_getter)} uncallable.')
    assert isinstance(other_pith_name_label, str), (
        f'{repr(other_pith_name_label)} not string.')

    # ....................{ VALIDATE                       }....................
    # Validate basic sanity.

    # If this hint does *NOT* directly annotate a callable parameter or return,
    # raise an exception. By definition, PEP 612-compliant parameter
    # specification variadic positional and keyword parameter type hints *MUST*
    # directly annotate the variadic positional and keyword parameter
    # (respectively) of a callable.
    if decor_meta is None:
        raise BeartypeDecorHintPep612Exception(
            f'{exception_prefix}PEP 612 "ParamSpec" '
            f'variadic {pith_name_label} parameter '
            f'type hint {repr(hint)} erroneously subscripts '
            f'a parent type hint as a nested child type hint rather than '
            f'directly annotating '
            f'variadic {pith_name_label} parameter "{pith_name_syntax}" as a '
            f'root type hint'
            f'{_get_pep612_exception_message_suffix()}'
        )
    # Else, this hint directly annotates a callable parameter or return.

    # ....................{ LOCALS                         }....................
    # Callable directly annotated by this hint, localized for readability.
    func = decor_meta.func_wrappee_wrappee

    # Metadata describing the other variadic parameter accepted by that callable
    # if that callable accepts the other variadic parameter *OR* "None"
    # otherwise (i.e., if that callable accepts *NO* such parameter).
    other_arg_meta = other_pith_name_getter(
        func=func,
        is_unwrap=False,  # <-- "func" is already a fully unwrapped callable
        exception_cls=BeartypeDecorHintPep612Exception,
        exception_prefix=exception_prefix,
    )
    # print(f'other_arg_meta: {other_arg_meta}')

    # ....................{ VALIDATE ~ this                }....................
    # Validate the passed variadic positional or keyword parameter type hint.

    # If this hint directly annotates a callable parameter or return that is
    # *NOT* a variadic positional or keyword parameter, raise an exception.
    if arg_kind is not arg_kind_expected:
        raise BeartypeDecorHintPep612Exception(
            f'{exception_prefix}PEP 612 "ParamSpec" '
            f'variadic {pith_name_label} parameter '
            f'type hint {repr(hint)} erroneously annotates '
            f'parameter "{pith_name}" rather than '
            f'variadic {pith_name_label} parameter "{pith_name_syntax}"'
            f'{_get_pep612_exception_message_suffix(func=func)}'
        )
    # Else, this hint directly annotates a variadic positional or keyword
    # parameter.

    # ....................{ VALIDATE ~ other               }....................
    # Validate the other variadic positional or keyword parameter type hint
    # required by PEP 612 to be paired with the passed hint.

    # If that callable accepts *NO* other variadic parameter, raise an
    # exception.
    if other_arg_meta is None:
        raise BeartypeDecorHintPep612Exception(
            f'{exception_prefix}PEP 612 "ParamSpec" '
            f'variadic {pith_name_label} parameter '
            f'type hint {repr(hint)} not paired with '
            f'corresponding PEP 612 "ParamSpec" '
            f'variadic {other_pith_name_label} parameter '
            f'type hint {repr(other_hint)} (i.e., '
            f'that callable accepts no '
            f'variadic {other_pith_name_label} parameter)'
            f'{_get_pep612_exception_message_suffix(func=func)}'
        )
    # Else, that callable accepts the other variadic parameter.

    # Name of the other variadic parameter.
    other_arg_name = other_arg_meta[ARG_META_INDEX_NAME]

    # Type hint subscripting the other variadic parameter if any *OR* the
    # sentinel placeholder otherwise.
    other_arg_hint = decor_meta.func_annotations_get(
        other_arg_name, SENTINEL)

    # If the other variadic parameter is unannotated, raise an exception.
    if other_arg_hint is SENTINEL:
        raise BeartypeDecorHintPep612Exception(
            f'{exception_prefix}PEP 612 "ParamSpec" '
            f'variadic {pith_name_label} parameter '
            f'type hint {repr(hint)} paired with unannotated '
            f'variadic {other_pith_name_label} parameter (i.e., '
            f'variadic {other_pith_name_label} parameter annotated by no '
            f'corresponding PEP 612 "ParamSpec" '
            f'variadic {other_pith_name_label} parameter '
            f'type hint {repr(other_hint)})'
            f'{_get_pep612_exception_message_suffix(func=func)}'
        )
    # Else, the other variadic parameter is annotated by a hint.

    # Sign uniquely identifying the hint annotating other variadic parameter.
    other_arg_hint_sign = get_hint_pep_sign_or_none(other_arg_hint)  # pyright: ignore

    # If the hint annotating the other variadic parameter is *NOT* the other
    # variadic positional or keyword parameter hint required by PEP
    # 612 to be paired with the passed hint, raise an exception.
    if other_arg_hint_sign is not other_hint_sign:
        raise BeartypeDecorHintPep612Exception(
            f'{exception_prefix}PEP 612 "ParamSpec" '
            f'variadic {pith_name_label} parameter '
            f'type hint {repr(hint)} paired with '
            f'variadic {other_pith_name_label} parameter erroneously annotated '
            f'by type hint {repr(other_arg_hint)} rather than '
            f'corresponding PEP 612 "ParamSpec" '
            f'variadic {other_pith_name_label} parameter '
            f'type hint {repr(other_hint)}'
            f'{_get_pep612_exception_message_suffix(func=func)}'
        )
    # Else, the type hint annotating the other variadic parameter is the
    # other variadic positional or keyword parameter type hint required by PEP
    # 612 to be paired with the passed hint, raise an exception.

    # Reduce *ALL* PEP 612 type hints to an arbitrary ignorable type hint.
    return object

# ....................{ PRIVATE ~ constants                }....................
def _get_pep612_exception_message_suffix(
    func: Optional[Callable] = None) -> str:
    '''
    Human-readable substring suffixing *all*
    :class:`.BeartypeDecorHintPep612Exception` messages raised by the
    :func:`._reduce_hint_pep612_args_or_kwargs` reducer.

    Parameters
    ----------
    func : Optional[Callable], optional
        Either:

        * If that reducer was passed a callable, that callable.
        * Else, :data:`None`.

        Defaults to :data:`None`.
    '''

    # Unqualified name of the passed callable if a callable was passed *OR* a
    # placeholder name otherwise.
    func_name = (
        func.__name__
        if func is not None else
        'big_chad_energy'  # <-- lol... ok. so, it's not actually that funny.
    )

    # Return the expected human-readable substring.
    return (
        f': e.g.,\n'
        f"\t# This is what PEP 612 wants. It's a big ask, but please try.\n"
        f'\tdef {func_name}(*args: P.args, **kwargs: P.kwargs): ...\n'
        f'You can do this. PEP 612 believes in you. '
        f'More importantly, so do we.'
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **callable factories** (i.e., low-level functions dynamically
creating and returning new in-memory callables).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilTypeException
from beartype.typing import Optional
from beartype._cave._cavemap import NoneTypeOr
from beartype._data.typing.datatyping import (
    LexicalScope,
    TupleTypes,
    TypeException,
)
from beartype._data.kind.datakindmap import FROZENDICT_EMPTY
from beartype._util.text.utiltextidentifier import die_unless_identifier

# ....................{ MAKERS                             }....................
def make_type(
    # Mandatory arguments.
    type_name: str,

    # Optional arguments.
    type_module_name: Optional[str] = None,
    type_bases: TupleTypes = (),
    type_scope: LexicalScope = FROZENDICT_EMPTY,
    type_doc: Optional[str] = None,
    exception_cls: TypeException = _BeartypeUtilTypeException,
    exception_prefix: str = '',
) -> type:
    '''
    Dynamically create and return a new class with the passed name subclassing
    all passed base classes and defined by the passed class scope.

    Parameters
    ----------
    type_name : str
        Name of the class to be created.
    type_module_name : Optional[str]
        Fully-qualified name of the module declaring this class. Defaults to
        :data:`None`, in which case this class remains undeclared by any module.
    type_bases : Tuple[type, ...]
        Tuple of all base classes to be inherited by this class. Defaults to
        the empty tuple, equivalent to the 1-tuple ``(object,)`` inheriting this
        class from only the root base class :class:`object` of all classes.
    type_scope : Optional[Dict[str, Any]]
        Dictionary mapping from the name to value of each **class-scoped
        attribute** (i.e., method, variable) to be defined by this class.
        Defaults to the empty dictionary, equivalent to declaring a class with
        the trivial body ``pass``.
    type_doc : Optional[str]
        Human-readable docstring documenting this class. Defaults to
        :data:`None`, in which case this class remains undocumented.
    exception_cls : Type[Exception], optional
        Type of exception to raise in the event of a fatal error. Defaults to
        :exc:`._BeartypeUtilTypeException`.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this object in the
        exception message. Defaults to the empty string.

    Returns
    -------
    type
        Class with this name subclassing these base classes and defined by this
        class scope.

    Raises
    ------
    exception_cls
        If either:

        * The passed classname is empty.
        * The passed classname is non-empty but *not* a valid unqualified Python
          identifier.
    '''
    assert isinstance(type_name, str), f'{repr(type_name)} not string.'
    assert isinstance(type_module_name, NoneTypeOr[str]), (
        f'{repr(type_module_name)} neither string nor "None".')
    assert isinstance(type_bases, tuple), f'{repr(type_bases)} not tuple.'
    assert isinstance(type_scope, dict), f'{repr(type_scope)} not dictionary.'
    assert isinstance(type_doc, NoneTypeOr[str]), (
        f'{repr(type_doc)} neither string nor "None".')

    # If this classname is *NOT* a valid unqualified Python identifier, raise an
    # exception. Insanely, the builtin type.__init__() constructor silently
    # allows this classname to be invalid -- despite the resulting class
    # violating sanity and normative standards. Note that invalid names include:
    # * The empty string.
    # * An invalid Python identifier.
    # * A valid fully-qualified Python identifier.
    if not type_name.isidentifier():
        raise exception_cls(
            f'{exception_prefix}class name {repr(type_name)} invalid.')
    # Else, this classname is a valid unqualified Python identifier.

    # Thank you, bizarre 3-parameter variant of the type.__init__() constructor.
    cls = type(type_name, type_bases, type_scope)

    # If this class has a module name...
    if type_module_name is not None:
        # If this module name is *NOT* a valid Python identifier, raise an
        # exception.
        die_unless_identifier(
            text=type_module_name,
            exception_cls=exception_cls,
            exception_prefix='Class module name ',
        )
        # Else, this module name is a valid Python identifier.

        # Set the module name of this class.
        cls.__module__ = type_module_name
    # Else, this class has *NO* module name.

    # If documenting this class, do so.
    if type_doc is not None:
        cls.__doc__ = type_doc
    # Else, this class is undocumented.

    # Return this class.
    return cls



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype decorator **type-checking expression magic** (i.e., global string
constants embedded in the implementations of boolean expressions type-checking
arbitrary objects against arbitrary PEP-compliant type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.error.dataerrmagic import EXCEPTION_PLACEHOLDER
# from itertools import count

# ....................{ EXCEPTION                          }....................
EXCEPTION_PREFIX_FUNC_WRAPPER_LOCAL = (
    f'{EXCEPTION_PLACEHOLDER}wrapper parameter ')
'''
Human-readable substring describing a new wrapper parameter required by the
current root type hint in exception messages.
'''


EXCEPTION_PREFIX_HINT = f'{EXCEPTION_PLACEHOLDER}type hint '
'''
Human-readable substring describing the current root type hint generically
(i.e., agnostic of the specific PEP standard to which this hint conforms) in
exception messages.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Callable parameter length getter utility unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.func.arg.utilfuncarglen` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_get_func_args_len_flexible() -> None:
    '''
    Test the
    :func:`beartype._util.func.arg.utilfuncarglen.get_func_args_flexible_len`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallableException
    from beartype._util.func.arg.utilfuncarglen import (
        get_func_args_flexible_len)
    from beartype_test.a00_unit.data.data_type import (
        ClassCallable,
        function_partial,
        function_partial_bad,
    )
    from beartype_test.a00_unit.data.func.data_func import (
        func_args_0,
        func_args_1_flex_mandatory,
        func_args_1_varpos,
        func_args_2_flex_mandatory,
        func_args_0_wrapped,
        func_args_1_flex_mandatory_wrapped,
        func_args_1_varpos_wrapped,
        func_args_2_flex_mandatory_wrapped,
    )
    from pytest import raises

    # ....................{ LOCALS                         }....................
    # Callable object whose class defines the __call__() dunder method.
    callable_object = ClassCallable()

    # ....................{ PASS                           }....................
    # Assert this getter returns the expected lengths of unwrapped callables.
    assert get_func_args_flexible_len(func_args_0) == 0
    assert get_func_args_flexible_len(func_args_1_varpos) == 0
    assert get_func_args_flexible_len(func_args_1_flex_mandatory) == 1
    assert get_func_args_flexible_len(func_args_2_flex_mandatory) == 2

    # Assert this getter returns the expected lengths of wrapped callables.
    assert get_func_args_flexible_len(func_args_0_wrapped) == 0
    assert get_func_args_flexible_len(func_args_1_varpos_wrapped) == 0
    assert get_func_args_flexible_len(func_args_1_flex_mandatory_wrapped) == 1
    assert get_func_args_flexible_len(func_args_2_flex_mandatory_wrapped) == 2

    # Assert this getter returns the expected length of a partial callable.
    assert get_func_args_flexible_len(function_partial) == 0

    # Assert this getter returns the expected length of a callable object.
    assert get_func_args_flexible_len(callable_object) == 0

    # Assert this getter returns 0 when passed a wrapped callable and an option
    # disabling callable unwrapping.
    assert get_func_args_flexible_len(
        func_args_0_wrapped, is_unwrap=False) == 0
    assert get_func_args_flexible_len(
        func_args_1_varpos_wrapped, is_unwrap=False) == 0
    assert get_func_args_flexible_len(
        func_args_1_flex_mandatory_wrapped, is_unwrap=False) == 0
    assert get_func_args_flexible_len(
        func_args_2_flex_mandatory_wrapped, is_unwrap=False) == 0

    # ....................{ FAIL                           }....................
    # Assert this getter raises the expected exception when passed an uncallable
    # object.
    with raises(_BeartypeUtilCallableException):
        get_func_args_flexible_len('Following his eager soul, the wanderer')

    # Assert this getter raises the expected exception when passed a C-based
    # callable.
    with raises(_BeartypeUtilCallableException):
        get_func_args_flexible_len(iter)

    # Assert this getter raises the expected exception when passed an invalid
    # partial callable passing more parameters than the underlying function
    # actually accepts.
    with raises(_BeartypeUtilCallableException):
        get_func_args_flexible_len(function_partial_bad)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **callable parameter tester utilities** (i.e., low-level functions
validating and testing various kinds of parameters accepted by arbitrary
callables).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilCallableException
from beartype._util.func.arg.utilfuncargiter import (
    ARG_META_INDEX_NAME,
)
from beartype._util.func.arg.utilfuncarglen import (
    ARGS_LENS_INDEX_VAR_POS,
    ARGS_LENS_INDEX_VAR_KW,
    get_func_args_flexible_len,
    get_func_args_len,
    get_func_args_lens,
    get_func_args_nonvariadic_len,
)
from beartype._data.typing.datatyping import (
    Codeobjable,
    TypeException,
)

# ....................{ RAISERS                            }....................
def die_unless_func_args_len_flexible_equal(
    # Mandatory parameters.
    func: Codeobjable,
    func_args_len_flexible: int,

    # Optional parameters.
    is_unwrap: bool = True,
    exception_cls: TypeException = _BeartypeUtilCallableException,
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception unless the passed pure-Python callable accepts the passed
    number of **flexible parameters** (i.e., parameters passable as either
    positional or keyword arguments).

    Parameters
    ----------
    func : Codeobjable
        Pure-Python callable, frame, or code object to be inspected.
    func_args_len_flexible : int
        Number of flexible parameters to validate this callable as accepting.
    is_unwrap: bool, default: True
        :data:`True` only if this validator implicitly calls the
        :func:`beartype._util.func.utilfuncwrap.unwrap_func_all` function to
        unwrap this possibly higher-level wrapper into its possibly lowest-level
        wrappee *before* returning the code object of that wrappee. Note that
        doing so incurs worst-case time complexity :math:`O(n)` for :math:`n`
        the number of lower-level wrappees wrapped by this wrapper. Defaults to
        :data:`True` for robustness. Why? Because this validator *must* always
        introspect lowest-level wrappees rather than higher-level wrappers. The
        latter typically do *not* accurately replicate the signatures of the
        former. In particular, decorator wrappers typically wrap decorated
        callables with variadic positional and keyword parameters (e.g., ``def
        _decorator_wrapper(*args, **kwargs)``). Since neither constitutes a
        flexible parameter, this validator raises an exception when passed such
        a wrapper with this boolean set to :data:`False`. Only set this boolean
        to :data:`False` if you pretend to know what you're doing.
    exception_cls : type[Exception], default: _BeartypeUtilCallableException
        Type of exception to be raised in the event of a fatal error. Defaults
        to :class:`._BeartypeUtilCallableException`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Raises
    ------
    exception_cls
        If this callable either:

        * Is *not* callable.
        * Is callable but is *not* pure-Python.
        * Is a pure-Python callable accepting either more or less than this
          Number of flexible parameters.
    '''
    assert isinstance(func_args_len_flexible, int)

    # Number of flexible parameters accepted by this callable.
    func_args_len_flexible_actual = get_func_args_flexible_len(
        func=func,
        is_unwrap=is_unwrap,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )

    # If this callable accepts more or less than this number of flexible
    # parameters, raise an exception.
    if func_args_len_flexible_actual != func_args_len_flexible:
        assert isinstance(exception_cls, type), (
            f'{repr(exception_cls)} not exception type.')
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        raise exception_cls(
            f'{exception_prefix}callable {repr(func)} flexible argument count '
            f'{func_args_len_flexible_actual} != {func_args_len_flexible} '
            f'(i.e., {repr(func)} accepts {func_args_len_flexible_actual} '
            f'rather than {func_args_len_flexible} positional and/or keyword '
            f'parameters).'
        )
    # Else, this callable accepts exactly this number of flexible parameters.


#FIXME: Uncomment as needed.
# def die_unless_func_argless(
#     # Mandatory parameters.
#     func: Codeobjable,
#
#     # Optional parameters.
#     func_label: str = 'Callable',
#     exception_cls: Type[Exception] = _BeartypeUtilCallableException,
# ) -> None:
#     '''
#     Raise an exception unless the passed pure-Python callable is
#     **argumentless** (i.e., accepts *no* arguments).
#
#     Parameters
#     ----------
#     func : Codeobjable
#         Pure-Python callable, frame, or code object to be inspected.
#     func_label : str, optional
#         Human-readable label describing this callable in exception messages
#         raised by this validator. Defaults to ``'Callable'``.
#     exception_cls : type, optional
#         Type of exception to be raised if this callable is neither a
#         pure-Python function nor method. Defaults to
#         :class:`_BeartypeUtilCallableException`.
#
#     Raises
#     ----------
#     exception_cls
#         If this callable either:
#
#         * Is *not* callable.
#         * Is callable but is *not* pure-Python.
#         * Is a pure-Python callable accepting one or more parameters.
#     '''
#
#     # If this callable accepts one or more arguments, raise an exception.
#     if is_func_argless(
#         func=func, func_label=func_label, exception_cls=exception_cls):
#         assert isinstance(func_label, str), f'{repr(func_label)} not string.'
#         assert isinstance(exception_cls, type), (
#             f'{repr(exception_cls)} not class.')
#
#         raise exception_cls(
#             f'{func_label} {repr(func)} not argumentless '
#             f'(i.e., accepts one or more arguments).'
#         )

# ....................{ TESTERS ~ kind                     }....................
def is_func_argless(
    # Mandatory parameters.
    func: Codeobjable,

    # Optional parameters.
    is_unwrap: bool = False,
    exception_cls: TypeException = _BeartypeUtilCallableException,
    exception_prefix: str = '',
) -> bool:
    '''
    :data:`True` only if the passed pure-Python callable is **argumentless**
    (i.e., accepts *no* arguments).

    Parameters
    ----------
    func : Codeobjable
        Pure-Python callable, frame, or code object to be inspected.
    is_unwrap: bool, optional
        :data:`True` only if this getter implicitly calls the
        :func:`beartype._util.func.utilfuncwrap.unwrap_func_all` function.
        Defaults to :data:`False` to avoid confusion with decorator wrappers,
        which almost always accept variadic arguments despite the callable they
        wrap *not* accepting such arguments. See also
        :func:`beartype._util.func.utilfunccodeobj.get_func_codeobj`.
    exception_cls : type, optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :class:`._BeartypeUtilCallableException`.
    exception_prefix : str, optional
        Human-readable label prefixing the message of any exception raised in
        the event of a fatal error. Defaults to the empty string.

    Returns
    -------
    bool
        :data:`True` only if the passed callable accepts *no* arguments.

    Raises
    ------
    exception_cls
         If the passed callable is *not* pure-Python.
    '''

    # Return true only if the passed callable accepts *NO* parameters.
    return not get_func_args_len(
        func=func,
        is_unwrap=is_unwrap,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )

# ....................{ TESTERS ~ kind : non-variadic      }....................
#FIXME: Unit test us up, please.
def is_func_arg_nonvariadic(*args, **kwargs) -> bool:
    '''
    :data:`True` only if the passed pure-Python callable accepts any
    **non-variadic parameters** (i.e., one or more positional, positional-only,
    keyword, or keyword-only arguments).

    Parameters
    ----------
    All parameters are passed as is to the lower-level
    :func:`beartype._util.func.utilfunccodeobj.get_func_codeobj` getter.

    Returns
    -------
    bool
        :data:`True` only if that callable accepts any non-variadic parameters.

    Raises
    ------
    exception_cls
         If that callable is *not* pure-Python.
    '''

    # Return true only if this callable accepts any non-variadic parameters.
    return bool(get_func_args_nonvariadic_len(*args, **kwargs))

# ....................{ TESTERS ~ kind : variadic          }....................
def is_func_arg_variadic(*args, **kwargs) -> bool:
    '''
    :data:`True` only if the passed pure-Python callable accepts any **variadic
    parameters** (i.e., either a variadic positional argument (e.g.,
    ``*args``) *or* a variadic keyword argument (e.g., ``**kwargs``)).

    Parameters
    ----------
    All parameters are passed as is to the lower-level
    :func:`beartype._util.func.utilfunccodeobj.get_func_codeobj` getter.

    Returns
    -------
    bool
        :data:`True` only if that callable accepts either:

        * Variadic positional arguments (e.g., ``*args``).
        * Variadic keyword arguments (e.g., ``**kwargs``).

    Raises
    ------
    exception_cls
         If that callable is *not* pure-Python.
    '''

    # Number of various kinds of parameters accepted by that callable.
    func_args_lens = get_func_args_lens(*args, **kwargs)

    # Return true only if this callable accepts any variadic argument.
    return bool(
        func_args_lens[ARGS_LENS_INDEX_VAR_POS] +
        func_args_lens[ARGS_LENS_INDEX_VAR_KW]
    )


def is_func_arg_variadic_positional(*args, **kwargs) -> bool:
    '''
    :data:`True` only if the passed pure-Python callable accepts a variadic
    positional argument (e.g., ``*args``).

    Parameters
    ----------
    All parameters are passed as is to the lower-level
    :func:`beartype._util.func.utilfunccodeobj.get_func_codeobj` getter.

    Returns
    -------
    bool
        :data:`True` only if the passed callable accepts a variadic positional
        argument.

    Raises
    ------
    exception_cls
         If the passed callable is *not* pure-Python.
    '''

    # Number of various kinds of parameters accepted by that callable.
    func_args_lens = get_func_args_lens(*args, **kwargs)

    # Return true only if this callable accepts a variadic positional argument.
    return func_args_lens[ARGS_LENS_INDEX_VAR_POS]  # type: ignore[return-value]


def is_func_arg_variadic_keyword(*args, **kwargs) -> bool:
    '''
    :data:`True` only if the passed pure-Python callable accepts a variadic
    keyword argument (e.g., ``**kwargs``).

    Parameters
    ----------
    All parameters are passed as is to the lower-level
    :func:`beartype._util.func.utilfunccodeobj.get_func_codeobj` getter.

    Returns
    -------
    bool
        :data:`True` only if the passed callable accepts a variadic keyword
        argument.

    Raises
    ------
    exception_cls
         If the passed callable is *not* pure-Python.
    '''

    # Number of various kinds of parameters accepted by that callable.
    func_args_lens = get_func_args_lens(*args, **kwargs)

    # Return true only if this callable accepts a variadic keyword argument.
    return func_args_lens[ARGS_LENS_INDEX_VAR_KW]  # type: ignore[return-value]

# ....................{ TESTERS ~ name                     }....................
def is_func_arg_name(arg_name: str, *args, **kwargs) -> bool:
    '''
    :data:`True` only if the passed pure-Python callable accepts an argument
    with the passed name.

    Parameters
    ----------
    arg_name : str
        Name of the argument to be searched for.

    All remaining parameters are passed as is to the lower-level
    :func:`beartype._util.func.utilfunccodeobj.get_func_codeobj` getter.

    Returns
    -------
    bool
        :data:`True` only if that callable accepts an argument with this name.

    Raises
    ------
    exception_cls
         If the passed callable is *not* pure-Python.
    '''
    assert isinstance(arg_name, str), f'{repr(arg_name)} not string.'

    # Avoid circular import dependencies.
    from beartype._util.func.arg.utilfuncargget import get_func_arg_names

    # Tuple of the names of all parameters accepted by the passed callable.
    arg_names = get_func_arg_names(*args, **kwargs)

    # Return true only if the passed name is in this tuple.
    #
    # Technically, this search has linear-time complexity "O(n)" for "n" the
    # number of parameters accepted by that callable. Pragmatically, "n" is
    # typically quite small; ergo, the real-world constants associated with this
    # complexity are also quite small. Although this search *COULD* be trivially
    # reduced to constant-time complexity "O(1)" by coercing this tuple into a
    # set, doing so would also substantially increase the real-world constants
    # associated with that complexity. In the absence of compelling profilings,
    # we currently prefer to abide by Keep It Simple Stupid (KISS). Tuple wins!
    return arg_name in arg_names


def is_func_arg_name_variadic_positional(
    arg_name: str, *args, **kwargs) -> bool:
    '''
    :data:`True` only if the passed pure-Python callable accepts a variadic
    positional parameter with the passed name.

    Caveats
    -------
    **This tester exhibits worst-case linear-time complexity** :math:`O(n)`
    **for** :math:`O(n)` **the total number of parameters accepted by that
    callable** due to unavoidably performing a linear search for a variadic
    positional parameter with this name is this callable's parameter list. This
    tester should thus be called sparingly.

    Parameters
    ----------
    arg_name : str
        Name of the argument to be searched for.

    All remaining parameters are passed as is to the lower-level
    :func:`beartype._util.func.utilfunccodeobj.get_func_codeobj` getter.

    Returns
    -------
    bool
        :data:`True` only if that callable accepts a variadic positional
        argument with this name.

    Raises
    ------
    exception_cls
         If the passed callable is *not* pure-Python.
    '''

    # Avoid circular import dependencies.
    from beartype._util.func.arg.utilfuncargget import (
        get_func_arg_meta_variadic_positional_or_none)

    # If the passed callable accepts *NO* parameter with the passed name,
    # immediately return false.
    if not is_func_arg_name(*args, arg_name=arg_name, **kwargs): # type: ignore[misc]
        return False
    # Else, that callable accepts a parameter with the passed name. In this
    # case, further testing is warranted.

    # Metadata describing the variadic positional parameter accepted by that
    # callable if that callable accepts a variadic positional parameter *OR*
    # "None" otherwise (i.e., if that callable accepts *NO* such parameter).
    arg_meta = get_func_arg_meta_variadic_positional_or_none(*args, **kwargs)

    # Return true only if...
    return (
        # That callable accepts a variadic positional parameter *AND*...
        arg_meta is not None and
        # The name of this parameter is the passed name.
        arg_meta[ARG_META_INDEX_NAME] == arg_name
    )


def is_func_arg_name_variadic_keyword(
    arg_name: str, *args, **kwargs) -> bool:
    '''
    :data:`True` only if the passed pure-Python callable accepts a variadic
    keyword parameter with the passed name.

    Caveats
    -------
    **This tester exhibits worst-case linear-time complexity** :math:`O(n)`
    **for** :math:`O(n)` **the total number of parameters accepted by that
    callable** due to unavoidably performing a linear search for a variadic
    keyword parameter with this name is this callable's parameter list. This
    tester should thus be called sparingly.

    Parameters
    ----------
    arg_name : str
        Name of the argument to be searched for.

    All remaining parameters are passed as is to the lower-level
    :func:`beartype._util.func.utilfunccodeobj.get_func_codeobj` getter.

    Returns
    -------
    bool
        :data:`True` only if that callable accepts a variadic keyword parameter
        with this name.

    Raises
    ------
    exception_cls
         If the passed callable is *not* pure-Python.
    '''

    # Avoid circular import dependencies.
    from beartype._util.func.arg.utilfuncargget import (
        get_func_arg_meta_variadic_keyword_or_none)

    # If the passed callable accepts *NO* parameter with the passed name,
    # immediately return false.
    if not is_func_arg_name(*args, arg_name=arg_name, **kwargs): # type: ignore[misc]
        return False
    # Else, that callable accepts a parameter with the passed name. In this
    # case, further testing is warranted.

    # Metadata describing the variadic keyword parameter accepted by that
    # callable if that callable accepts a variadic keyword parameter *OR* "None"
    # otherwise (i.e., if that callable accepts *NO* such parameter).
    arg_meta = get_func_arg_meta_variadic_keyword_or_none(*args, **kwargs)

    # Return true only if...
    return (
        # That callable accepts a variadic keyword parameter *AND*...
        arg_meta is not None and
        # The name of this parameter is the passed name.
        arg_meta[ARG_META_INDEX_NAME] == arg_name
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Unmemoized beartype non-type decorators** (i.e., low-level decorators
decorating *all* types of decoratable objects except classes, which the sibling
:mod:`beartype._decor._type.decortype` submodule handles, on behalf of the parent
:mod:`beartype._decor.decorcore` submodule).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import (
    BeartypeDecorWrappeeException,
    BeartypeDecorWrapperException,
)
from beartype.typing import (
    Optional,
    no_type_check,
)
from beartype._check.metadata.metadecor import (
    cull_beartype_call,
    make_beartype_call,
)
from beartype._conf.confmain import BeartypeConf
from beartype._conf.confenum import BeartypeStrategy
from beartype._decor._nontype._decornontypemap import (
    MODULE_TO_TYPE_NAME_TO_BEARTYPE_DECORATOR_get,
    MODULE_TO_SUPERTYPE_NAME_TO_BEARTYPE_DECORATOR_get,
)
from beartype._data.typing.datatyping import BeartypeableT
from beartype._decor._nontype._wrap.wrapmain import generate_code
from beartype._util.bear.utilbearblack import is_object_blacklisted
from beartype._util.bear.utilbearfunc import (
    is_func_unbeartypeable,
    set_func_beartyped,
)
from beartype._util.api.standard.utilcontextlib import (
    get_func_contextlib_contextmanager_or_none)
from beartype._util.api.standard.utilfunctools import (
    beartype_functools_lru_cache,
    is_func_functools_lru_cache,
)
from beartype._util.func.utilfuncmake import make_func
from beartype._util.func.utilfunctest import (
    is_func_codeobjable,
    is_func_wrapper,
)
from beartype._util.func.utilfuncwrap import unwrap_func_once
from beartype._util.module.utilmodget import get_object_module_name_or_none
from beartype._util.text.utiltextrepr import represent_object
from collections.abc import Callable

# ....................{ DECORATORS                         }....................
def beartype_nontype(obj: BeartypeableT, **kwargs) -> BeartypeableT:
    '''
    Decorate the passed **non-class beartypeable** (i.e., caller-defined object
    that may be decorated by the :func:`beartype.beartype` decorator but is
    *not* a class) with dynamically generated type-checking.

    Parameters
    ----------
    obj : BeartypeableT
        Non-class beartypeable to be decorated.

    All remaining keyword parameters are passed as is to a lower-level decorator
    defined by this submodule (e.g., :func:`.beartype_func`).

    Returns
    -------
    BeartypeableT
        New pure-Python callable wrapping this beartypeable with type-checking.
    '''

    # Validate that the passed object is *NOT* a class.
    assert not isinstance(obj, type), f'{repr(obj)} is class.'
    # print(f'Decorating non-type {repr(obj)} with type-checking...')
    # print(f'Non-type contents: {dir(obj)}')
    # print(f'{dir(obj.__code__)}')
    # print(f'{obj.__code__.co_filename}')
    # print(f'{obj.__code__.co_name}')
    # print(f'{obj.__code__.co_names}')
    # print(f'{obj.__code__.co_qualname}')

    # ....................{ PASS 1 ~ dispatch : O(1)       }....................
    # First-pass logic efficiently dispatching a beartype decorator unique to
    # the type of this object in O(1) constant-time dispatch -- including:
    # * Uncallable builtin method descriptors -- including property, class
    #   method, instance method, and static method objects. In this case,
    #   @beartype was listed above rather than below the builtin decorator
    #   generating this descriptor in the chain of decorators decorating this
    #   decorated callable. Although @beartype typically *MUST* decorate a
    #   callable directly, this edge case is sufficiently common *AND* trivial
    #   to resolve to warrant doing so. To do so, this conditional branch
    #   effectively reorders @beartype to be the first decorator decorating the
    #   pure-Python function underlying this method descriptor: e.g.,
    #       # This branch detects and reorders this edge case...
    #       class MuhClass(object):
    #           @beartype
    #           @classmethod
    #           def muh_classmethod(cls) -> None: pass
    #
    #       # ...to resemble this direct decoration instead.
    #       class MuhClass(object):
    #           @classmethod
    #           @beartype
    #           def muh_classmethod(cls) -> None: pass
    #
    #   Note that most but *NOT* all of these objects are uncallable.
    #   Regardless, *ALL* of these objects are unsuitable for direct decoration.
    #   Specifically:
    #   * Under Python < 3.10, *ALL* of these objects are uncallable.
    #   * Under Python >= 3.10:
    #     * Descriptors created by @classmethod and @property are uncallable.
    #     * Descriptors created by @staticmethod are technically callable but
    #       C-based and thus unsuitable for direct decoration.

    # Type of this object.
    obj_type = obj.__class__

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Synchronize with the "PHASE 2" heuristic implemented below.
    # Although these two heuristics are suspiciously similar, generalizing them
    # into a single utility function would only decrease efficiency and increase
    # complexity for no particularly good reason.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # Fully-qualified name of the module defining this type if any *OR* "None"
    # (i.e., if this type is only defined in-memory outside of a module).
    obj_type_module_name = get_object_module_name_or_none(obj_type)

    # If this type is defined by a module...
    if obj_type_module_name:
        # Dictionary mapping from the unqualified basename of each well-known
        # standard type of a @beartype-decoratable object defined by the module
        # defining this specific type to a decorator decorating that object with
        # dynamically generated type-checking if any *OR* "None" (i.e., if this
        # type is *NOT* defined by such a well-known dispatchable module).
        obj_type_name_to_beartype_decorator = (
            MODULE_TO_TYPE_NAME_TO_BEARTYPE_DECORATOR_get(
                obj_type_module_name))

        # If this type is defined by a well-known dispatchable module...
        if obj_type_name_to_beartype_decorator:
            # Decorator decorating this object with dynamically generated
            # type-checking if any *OR* "None" (i.e., if this type is *NOT* a
            # well-known type defined by this module).
            beartype_decorator = obj_type_name_to_beartype_decorator.get(
                obj_type.__name__)

            # If this type is dispatchable, trivially do so.
            if beartype_decorator:
                return beartype_decorator(obj, **kwargs)  # type: ignore[return-value]
            # Else, this type is *NOT* dispatchable.
        # Else, this type is *NOT* defined by a well-known dispatchable module.
    # Else, this type is defined in-memory outside a module.

    # ....................{ PASS 2 ~ dispatch : O(n)       }....................
    # Second-pass logic less efficiently dispatching a beartype decorator unique
    # to the method-resolution order (MRO) of this object in O(n) constant-time
    # dispatch -- including:
    # * Callable third-party objects requiring special handling -- including:
    #   * Click commands created by the @click.command() decorator.

    # Tuple of all superclasses of this object (including the type of this
    # object, which is of course its own superclass *AND* subclass, because set
    # theory just goes hard like that).
    #
    # Note that this includes the irrelevant root "object" superclass, which is
    # semantically meaningless and thus guaranteed to *NEVER* be matched below.
    # Although that superclass *COULD* be trivially sliced off (e.g., with an
    # assignment resembling "obj_bases = cls.__mro__[:-1]"), doing so only
    # uselessly consumes more time than it saves. So it goes.
    obj_bases = obj_type.__mro__

    # For each superclass of this object...
    for obj_base in obj_bases:
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # CAUTION: Synchronize with the "PHASE 1" heuristic implemented above.
        # Although these two heuristics are suspiciously similar, generalizing
        # them into a single utility function would only decrease efficiency and
        # increase complexity for no particularly good reason.
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        # Fully-qualified name of the module defining this superclass if any
        # *OR* "None" (i.e., if only defined in-memory outside of a module).
        obj_base_module_name = get_object_module_name_or_none(obj_base)

        # If this superclass is defined by a module...
        if obj_base_module_name:
            # Dictionary mapping from the unqualified basename of each
            # well-known superclass of a @beartype-decoratable object defined
            # by the module defining this specific type to a decorator
            # decorating that object with dynamically generated type-checking if
            # any *OR* "None" (i.e., if this superclass is *NOT* defined by such
            # a well-known dispatchable module).
            obj_base_name_to_beartype_decorator = (
                MODULE_TO_SUPERTYPE_NAME_TO_BEARTYPE_DECORATOR_get(
                    obj_base_module_name))

            # If this superclass is defined by a well-known dispatchable
            # module...
            if obj_base_name_to_beartype_decorator:
                # Decorator decorating this object with dynamically generated
                # type-checking if any *OR* "None" (i.e., if this superclass is
                # *NOT* a well-known superclass defined by this module).
                beartype_decorator = obj_base_name_to_beartype_decorator.get(  # type: ignore[assignment]
                    obj_base.__name__)

                # If this superclass is dispatchable, trivially do so.
                if beartype_decorator:
                    return beartype_decorator(obj, **kwargs)  # type: ignore[return-value]
                # Else, this superclass is *NOT* dispatchable.
            # Else, this superclass is *NOT* defined by a well-known
            # dispatchable module.
        # Else, this superclass is defined in-memory outside a module.

    # ....................{ PASS 3 ~ ad-hoc                }....................
    # Third-pass logic inefficiently dispatching a beartype decorator unique to
    # the type of this object with a series of ad-hoc heuristics, each of which
    # does *NOT* reduce to a trivial check against the fully-qualified name of
    # the type of this callable and thus *CANNOT* be integrated into the
    # efficient mapping-based O(1) dispatch employed above.

    # If this object is beartype-blacklisted (i.e., defined in a third-party
    # package or module that is hostile to runtime type-checking), silently
    # reduce to a noop and preserve this object as is -- even if this object is
    # uncallable. Of course, this is hardly ideal. But...
    #
    # Beartype didn't break it. Beartype can't fix it. Beartype ignores it!
    if is_object_blacklisted(obj):
        return obj
    # Else, this object is *NOT* beartype-blacklisted.
    #
    # If this object is uncallable, raise an exception.
    elif not callable(obj):
        raise BeartypeDecorWrappeeException(
            f'Uncallable {represent_object(obj)} not decoratable by @beartype.')
    # Else, this object is callable.
    #
    # If this object is *NOT* a pure-Python function, this object is a
    # pseudo-callable (i.e., arbitrary pure-Python *OR* C-based object whose
    # class defines the __call__() dunder method enabling this object to be
    # called like a standard callable). In this case, attempt to monkey-patch
    # runtime type-checking into this pseudo-callable by replacing the bound
    # method descriptor of the type of this object implementing the __call__()
    # dunder method with a comparable descriptor calling a @beartype-generated
    # runtime type-checking wrapper function. Go with it.
    elif not is_func_codeobjable(obj):
        return _beartype_pseudofunc(obj, **kwargs)  # type: ignore[return-value]
    # Else, this object is a pure-Python function.

    # Either:
    # * If this function is a "contextlib"-based isomorphic decorator closure
    #   (i.e., closure both created and returned by either the standard
    #   @contextlib.asynccontextmanager or @contextlib.contextmanager decorators
    #   where that closure isomorphically preserves both the number and types of
    #   all passed parameters and returns by accepting only a variadic positional
    #   argument and variadic keyword argument), that decorator.
    # * Else, "None".
    func_contextmanager = get_func_contextlib_contextmanager_or_none(obj)

    # If this function is a "contextlib"-based isomorphic decorator closure,
    # @beartype was listed above rather than below the "contextlib" decorator
    # creating and returning this closure in the chain of decorators decorating
    # this decorated callable.
    #
    # This is non-ideal, as the types of *ALL* objects created and returned by
    # "contextlib"-decorated context managers are private classes of the
    # "contextlib" module rather than the types implied by the type hints
    # originally annotating the returns of those context managers. If @beartype
    # failed to actively detect and intervene in this edge case, then runtime
    # type-checkers dynamically generated by @beartype for those managers would
    # erroneously raise type-checking violations after calling those managers
    # and detecting the apparent type violation: e.g.,
    #     >>> from beartype.typing import Iterator
    #     >>> from contextlib import contextmanager
    #     >>> @contextmanager
    #     ... def muh_context_manager() -> Iterator[None]: yield
    #     >>> type(muh_context_manager())
    #     <class 'contextlib._GeneratorContextManager'>  # <-- not an "Iterator"
    #
    # This conditional branch effectively reorders @beartype to be the first
    # decorator decorating the callable underlying this context manager,
    # preserving consistency between return types *AND* return type hints: e.g.,
    #     from beartype.typing import Iterator
    #     from contextlib import contextmanager
    #
    #     # This branch detects and reorders this edge case...
    #     @beartype
    #     @contextmanager
    #     def muh_contextmanager(cls) -> Iterator[None]: yield
    #
    #     # ...to resemble this direct decoration instead.
    #     @contextmanager
    #     @beartype
    #     def muh_contextmanager(cls) -> Iterator[None]: yield
    #
    # Note that detecting "contextlib"-based isomorphic decorator closures is
    # extremely non-trivial. Notably, this detection requires Python >= 3.11 and
    # silently fails under Python <= 3.10. It is what it is. Not our fault!
    if func_contextmanager is not None:
        return _beartype_func_contextlib_contextmanager(  # type: ignore[return-value]
            func=obj, func_contextmanager=func_contextmanager, **kwargs)
    # Else, that function is *NOT* a "contextlib"-based isomorphic decorator
    # closure. By elimination, that function *MUST* be a standard pure-Python
    # function.

    # Decorate that pure-Python function with runtime type-checking.
    return beartype_func(obj, **kwargs)  # type: ignore[return-value]


def beartype_func(
    # Mandatory parameters.
    func: BeartypeableT,
    conf: BeartypeConf,

    # Variadic keyword parameters.
    wrapper: Optional[Callable] = None,
    **kwargs
) -> BeartypeableT:
    '''
    Decorate the passed callable with dynamically generated type-checking.

    Parameters
    ----------
    func : BeartypeableT
        Callable to be decorated by :func:`beartype.beartype`.
    conf : BeartypeConf
        Beartype configuration configuring :func:`beartype.beartype` uniquely
        specific to this callable.
    wrapper : Optional[Callable]
        Wrapper callable to be unwrapped in the event that the callable to be
        unwrapped differs from the callable to be decorated. Typically, these
        two callables are the same. Edge cases in which these two callables
        differ include:

        * When ``wrapper`` is a **pseudo-callable** (i.e., otherwise uncallable
          object whose type renders that object callable by defining the
          ``__call__()`` dunder method) *and* ``func`` is that ``__call__()``
          dunder method. If that pseudo-callable wraps a lower-level callable,
          then that pseudo-callable (rather than ``__call__()`` dunder method)
          defines the ``__wrapped__`` instance variable providing that callable.

        Defaults to :data:`None`, in which case this parameter *actually*
        defaults to ``func``.

    All remaining keyword parameters are passed as is to the
    :meth:`beartype._check.metadata.metadecor.BeartypeDecorMeta.reinit` method.

    Returns
    -------
    BeartypeableT
        New pure-Python callable wrapping this callable with type-checking.
    '''

    # If the caller failed to pass a callable to be unwrapped, default that to
    # the callable to be type-checked.
    if wrapper is None:
        wrapper = func  # type: ignore[assignment]
    # Else, the caller passed a callable to be unwrapped. Preserve it up!

    # Validate all explicitly passed parameters.
    assert callable(func), f'{repr(func)} uncallable.'
    assert callable(wrapper), f'{repr(wrapper)} uncallable.'
    assert isinstance(conf, BeartypeConf), f'{repr(conf)} not configuration.'

    #FIXME: Uncomment to display all annotations in "pytest" tracebacks.
    # func_hints = func.__annotations__

    # If this configuration enables the no-time strategy performing *NO*
    # type-checking, monkey-patch that callable with the standard
    # @typing.no_type_check decorator detected below by the call to the
    # is_func_unbeartypeable() tester on all subsequent decorations passed the
    # same callable... Doing so prevents all subsequent decorations from
    # erroneously ignoring this previously applied no-time strategy.
    if conf.strategy is BeartypeStrategy.O0:
        no_type_check(func)  # pyright: ignore
    # Else, this configuration enables a positive-time strategy performing at
    # least the minimal amount of type-checking.

    # If the callable to be unwrapped is unbeartypeable (i.e., if this decorator
    # should preserve that callable as is rather than wrap that callable with
    # type-checking), silently reduce to the identity decorator.
    #
    # Note that this conditional implicitly handles the prior conditional! Ergo,
    # this conditional intentionally appears *AFTER* the prior conditional. :O
    if is_func_unbeartypeable(wrapper):  # type: ignore[arg-type]
        # print(f'Ignoring unbeartypeable callable {repr(func)}...')
        return func  # type: ignore[return-value]
    # Else, that callable is beartypeable. Let's do this, folks.

    # Beartype call metadata describing that callable.
    decor_meta = make_beartype_call(
        func=func, conf=conf, wrapper=wrapper, **kwargs)  # pyright: ignore
    # print(f'Decorating {repr(decor_meta)} with wrapper {repr(wrapper)}...')

    # Generate the raw string of Python statements implementing this wrapper.
    func_wrapper_code = generate_code(decor_meta)
    # print(f'func_wrapper_code: {func_wrapper_code}')

    # If that callable requires *NO* type-checking, silently reduce to a noop
    # and thus the identity decorator by returning that callable as is.
    if not func_wrapper_code:
        return func  # type: ignore[return-value]
    # Else, that callable requires type-checking. Let's *REALLY* do this, fam.

    # If the type hint dictionary associated with the decorated callable is
    # dirty (i.e., changed from the original "__annotations__" dunder dictionary
    # annotating that callable), register these changes in a manner compliant
    # with both PEP 649 and Python >= 3.14 *BEFORE* calling the make_func()
    # factory function below, which internally propagates these changes from the
    # "decor_meta.func_wrapper" callable into the created type-checking wrapper
    # function returned by the @beartype decorator. Look. It's complicated.
    decor_meta.set_func_annotations_if_dirty()

    # Function wrapping that callable with type-checking to be returned.
    #
    # For efficiency, this wrapper accesses *ONLY* local rather than global
    # attributes. The latter incur a minor performance penalty, since local
    # attributes take precedence over global attributes, implying all global
    # attributes are *ALWAYS* first looked up as local attributes before falling
    # back to being looked up as global attributes.
    func_wrapper = make_func(
        func_name=decor_meta.func_wrapper_name,
        func_code=func_wrapper_code,
        func_locals=decor_meta.func_wrapper_scope,
        func_wrapped=decor_meta.func_wrapper,  # pyright: ignore
        func_labeller=decor_meta.label_func_wrapper,
        is_debug=conf.is_debug,
        exception_cls=BeartypeDecorWrapperException,
    )

    # Declare this wrapper to be generated by @beartype, which tests for the
    # existence of this attribute above to avoid re-decorating callables
    # already decorated by @beartype by efficiently reducing to a noop.
    set_func_beartyped(func_wrapper)

    # Deinitialize this beartype call metadata.
    cull_beartype_call(decor_meta)

    # Return this wrapper.
    return func_wrapper  # type: ignore[return-value]

# ....................{ PRIVATE ~ decorators : pure-python }....................
def _beartype_func_contextlib_contextmanager(
    func: BeartypeableT,
    func_contextmanager: Callable,
    **kwargs
) -> BeartypeableT:
    '''
    Decorate the passed :mod:`contextlib`-based **isomorphic decorator closure**
    (i.e., closure both defined and returned by either the standard
    :func:`contextlib.asynccontextmanager` or :func:`contextlib.contextmanager`
    decorator where that closure isomorphically preserves both the number and
    types of all passed parameters and returns by accepting only a variadic
    positional argument and variadic keyword argument) with dynamically
    generated type-checking.

    Parameters
    ----------
    func : BeartypeableT
        Context manager to be decorated by :func:`beartype.beartype`.
    func_contextmanager : Callable
        Either:

        * If this context manager is a
          :func:`contextlib.asynccontextmanager`-based isomorphic decorator
          closure, :func:`contextlib.asynccontextmanager`.
        * Else, this context manager is a
          :func:`contextlib.contextmanager`-based isomorphic decorator closure.
          In this case, :func:`contextlib.contextmanager`.

    All remaining keyword parameters are passed as is to the lower-level
    :func:`.beartype_func` decorator internally called by this higher-level
    decorator on the pure-Python function encapsulated in this descriptor.

    Returns
    -------
    BeartypeableT
        New pure-Python callable wrapping this context manager with
        type-checking.
    '''
    assert callable(func_contextmanager), (
        f'{repr(func_contextmanager)} uncallable.')

    # Original pure-Python generator factory function decorated by either the
    # @contextlib.asynccontextmanager or @contextlib.contextmanager decorator.
    generator = unwrap_func_once(func)  # type: ignore[arg-type]

    # Decorate this generator factory function with type-checking.
    generator_checked = beartype_func(func=generator, **kwargs)

    # Re-decorate this generator factory function by the same decorator.
    generator_checked_contextmanager = func_contextmanager(generator_checked)

    # Return this context manager.
    return generator_checked_contextmanager  # type: ignore[return-value]


def _beartype_pseudofunc(pseudofunc: BeartypeableT, **kwargs) -> BeartypeableT:
    '''
    Monkey-patch the passed **pseudo-callable** (i.e., arbitrary pure-Python
    *or* C-based object whose class defines the ``__call__()`` dunder method
    enabling this object to be called like a standard callable) with dynamically
    generated type-checking.

    For each bound method descriptor encapsulating a method bound to this
    object, this function monkey-patches (i.e., replaces) that descriptor with a
    comparable descriptor calling a new :func:`beartype.beartype`-generated
    runtime type-checking wrapper function wrapping the original method.

    Parameters
    ----------
    pseudofunc : BeartypeableT
        Pseudo-callable to be monkey-patched by :func:`beartype.beartype`.

    All remaining keyword parameters are passed as is to the lower-level
    :func:`.beartype_func` decorator internally called by this higher-level
    decorator on the pure-Python function encapsulated in this descriptor.

    Returns
    -------
    BeartypeableT
        The object monkey-patched by :func:`beartype.beartype`.
    '''
    # print(f'@beartyping pseudo-callable {repr(pseudofunc)}...')

    # Bound __call__() dunder method bound to this object if this object defines
    # this method *OR* "None" otherwise.
    pseudofunc_call_boundmethod = getattr(pseudofunc, '__call__')

    # Unbound __call__() dunder method defined by the type of this object if
    # this type defines this method *OR* "None" otherwise.
    pseudofunc_call_type_method = getattr(pseudofunc.__class__, '__call__')

    # If this object does *NOT* define this method, this object is *NOT* a
    # pseudo-callable. In this case, raise an exception.
    #
    # Note this edge case should *NEVER* occur. By definition, this object has
    # already been validated to be callable. But this object is *NOT* a
    # pure-Python function. Since the only other category of callable in Python
    # is a pseudo-callable, this object *MUST* be a pseudo-callable. That said,
    # languages change; it's not inconceivable that Python could introduce yet
    # another kind of callable object under future versions.
    if pseudofunc_call_boundmethod is None:  # pragma: no cover
        raise BeartypeDecorWrappeeException(
            f'Callable {repr(pseudofunc)} not pseudo-callable object '
            f'(i.e., defines no bound __call__() dunder method).'
        )
    # Else, this object is a pseudo-callable.
    #
    # If this object does *NOT* define this method, this object is *NOT* a
    # pseudo-callable. In this case, raise an exception.
    elif pseudofunc_call_type_method is None:  # pragma: no cover
        raise BeartypeDecorWrappeeException(
            f'Callable {repr(pseudofunc)} type {repr(pseudofunc.__class__)} '
            f'not pseudo-callable object type '
            f'(i.e., defines no unbound __call__() dunder method).'
        )
    # Else, this object type is a pseudo-callable type.
    #
    # If this is a C-based @functools.lru_cache-memoized callable (i.e.,
    # low-level C-based callable object both created and returned by the
    # standard @functools.lru_cache decorator), @beartype was listed above
    # rather than below the @functools.lru_cache decorator creating and
    # returning this callable in the chain of decorators decorating this
    # decorated callable.
    #
    # This conditional branch effectively reorders @beartype to be the first
    # decorator decorating the pure-Python callable underlying this C-based
    # pseudo-callable: e.g.,
    #     from functools import lru_cache
    #
    #     # This branch detects and reorders this edge case...
    #     @beartype
    #     @lru_cache
    #     def muh_lru_cache() -> None: pass
    #
    #     # ...to resemble this direct decoration instead.
    #     @lru_cache
    #     @beartype
    #     def muh_lru_cache() -> None: pass
    elif is_func_functools_lru_cache(pseudofunc):
        # Return a new callable decorating that callable with type-checking.
        return beartype_functools_lru_cache(  # type: ignore
            pseudofunc=pseudofunc, **kwargs)  # pyright: ignore
    # Else, this is *NOT* a C-based @functools.lru_cache-memoized callable.
    #
    # If...
    elif (
        # This pseudo-callable object is a wrapper *AND*...
        is_func_wrapper(pseudofunc) and
        # This unbound __call__() dunder method is *NOT* a wrapper...
        not is_func_wrapper(pseudofunc_call_type_method)
    ):
        # print(f'Pseudo-callable wrapper {repr(pseudofunc)} identified!')

        # Transitively pass the optional "wrapper" parameter to the
        # BeartypeDecorMeta.reinit() method, ensuring that this pseudo-callable
        # wrapper object is correctly unwrapped.
        #
        # This edge case handles edge-case pseudo-callable wrapper objects
        # defined by popular third-party packages, including:
        # * The pseudo-callable wrapper objects created and returned by the
        #   @equinox.filter_jit wrapper. Although private and extremely
        #   non-trivial, the types of these objects vaguely resembles:
        #       class _JitWrapper(object):
        #           def __init__(self, func):
        #               self.__wrapped__ = func
        #
        #           def __call__(self, *args, **kwargs):
        #               return self.__wrapped__(*args, **kwargs)
        kwargs['wrapper'] = pseudofunc
    # Else, either this pseudo-callable object is not a wrapper *OR* this
    # unbound __call__() dunder method is already a wrapper.

    # Unbound __call__() dunder method runtime type-checking the original bound
    # __call__() dunder method of the passed pseudo-callable object.
    pseudofunc_call_type_method_checked = beartype_func(
        func=pseudofunc_call_boundmethod, **kwargs)
    return pseudofunc_call_type_method_checked



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **data subpackage cleaners** (i.e., low-level functions internally
called by higher-level tests and fixtures to recursively remove all bytecode
files from imported test-specific data subpackages).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ FIXTURES                           }....................
def data_claw_subpackage_cleaned(
    data_claw_subpackage: 'types.ModuleType') -> None:
    '''
    Context manager guaranteeably reverting *all* beartype import hooks
    transitively installed in the body of this context manager by the parent
    unit test implicitly invoking this fixture.

    Before doing so, this context manager permanently, silently, and recursively
    removes all **bytecode files** (i.e., pure-Python bytecode compiled to
    platform-dependent temporary files residing in temporary ``__pycache__/``
    subdirectories) of the passed test-specific subpackage and all
    subsubpackages of that subpackage regardless of depth.

    Parameters
    ----------
    data_claw_subpackage : types.ModuleType
        Previously imported subpackage containing one or more submodules
        expected to be hooked by :mod:`beartype.claw` import hooks. This
        subpackage should typically either be:

        * The :mod:`beartype_test.a00_unit.data.claw` subpackage.
        * The :mod:`beartype_test.a90_func.data.claw` subpackage.

    Raises
    ------
    beartype.roar.BeartypeClawHookException
        If one or more packages have been registered by a prior call to the
        :func:`beartype.claw._package.clawpkgmain.hook_packages` function.
    '''

    # Defer fixture-specific imports.
    from beartype.claw._package.clawpkgcontext import packages_trie_reverted

    # Recursively remove *ALL* previously compiled bytecode files from both this
    # subpackage *and* *ALL* subsubpackages of this subpackage.
    clean_data_subpackage(data_claw_subpackage)

    # With a context manager guaranteeably reverting *ALL* beartype import hooks
    # transitively installed in the body of this context manager, defer to the
    # parent unit test implicitly invoking this fixture.
    #
    # Note that it is no longer safe to unconditionally clear *ALL* beartype
    # import hooks. Previously, when beartype was less popular, it was. That is
    # exactly what we did. At some point, however, it would seem that one or
    # more third-party packages of unknown provenance imported at an unknown
    # time by our test suite began registering beartype import hooks against
    # Microsoft's ambiguously named "key_value" Python package: the official API
    # for their Azure cloud-hosting services. Since GitHub Actions literally run
    # on Azure, that package now appears to be type-checked by beartype *BEFORE*
    # the beartype test suite is even started under GitHub Actions-based
    # continuous integration (CI) workflows. Crazy, yet true.
    with packages_trie_reverted():
        yield

# ....................{ CLEANERS                           }....................
def clean_data_subpackage(data_subpackage: 'types.ModuleType') -> None:
    '''
    Permanently, silently, and recursively remove all **bytecode files** (i.e.,
    pure-Python bytecode compiled to platform-dependent temporary files residing
    in temporary ``__pycache__/`` subdirectories) of the passed test-specific
    subpackage and all subsubpackages of that subpackage regardless of depth.

    Parameters
    ----------
    datasubpackage : types.ModuleType
        Previously imported subpackage containing one or more data submodules.

    See Also
    --------
    :func:`beartype._util.path.utilpathremove.remove_package_bytecode_files`
        Further details.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer fixture-specific imports.
    from beartype._util.module.utilmodget import get_module_dir
    from beartype._util.path.utilpathremove import remove_package_bytecode_files

    # ....................{ PATHS                          }....................
    # Path encapsulating the absolute dirname of this data claw subpackage.
    data_subpackage_dir = get_module_dir(data_subpackage)
    # print(f'Clearing "{data_subpackage_dir}/" bytecode files...')

    # Recursively remove *ALL* previously compiled bytecode files from both this
    # subdirectory *and* *ALL* subsubdirectories of this subdirectory.
    remove_package_bytecode_files(data_subpackage_dir)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`593` unit tests.

This submodule unit tests :pep:`593` support implemented in the
:func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_decor_pep593() -> None:
    '''
    Test :pep:`593` support implemented in the :func:`beartype.beartype`
    decorator.

    Notably, this test exercises unsuccessful usage of the public
    :mod:`beartype.vale.Is` class when used to type hint callables decorated by
    the :func:`beartype.beartype` decorator.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeDecorHintPep593Exception
    from beartype.vale import Is
    from beartype._util.api.standard.utiltyping import get_typing_attrs
    from pytest import raises

    # ..................{ FAIL                               }..................
    # For the "Annotated" type hint factory exported by each typing module...
    for Annotated in get_typing_attrs('Annotated'):
        # Assert that @beartype raises the expected exception when decorating a
        # callable annotated by a type metahint whose first argument is a
        # beartype-specific data validator and whose second argument is a
        # beartype-agnostic object.
        with raises(BeartypeDecorHintPep593Exception):
            @beartype
            def volleyed_and_thundered(
                flashed_all_their_sabres_bare: Annotated[
                    str,
                    Is[lambda text: bool('Flashed as they turned in air')],
                    'Sabring the gunners there,',
                ]
            ) -> str:
                return flashed_all_their_sabres_bare + 'Charging an army, while'



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **platform tester** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.os.utilostest` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ tester                     }....................
def test_is_os_macos() -> None:
    '''
    Test the :func:`beartype._util.os.utilostest.is_os_macos` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.os.utilostest import is_os_macos

    # Assert this tester returns a boolean.
    IS_OS_MACOS = is_os_macos()
    assert isinstance(IS_OS_MACOS, bool)


def test_is_os_windows_vanilla() -> None:
    '''
    Test the :func:`beartype._util.os.utilostest.is_os_windows_vanilla` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.os.utilostest import is_os_windows_vanilla

    # Assert this tester returns a boolean.
    IS_OS_WINDOWS_VANILLA = is_os_windows_vanilla()
    assert isinstance(IS_OS_WINDOWS_VANILLA, bool)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`695`-compliant PEP-noncompliant type hint test data.

:pep:`695`-compliant type hints *mostly* indistinguishable from
PEP-noncompliant type hints include:

* :func:`typing.TypeAliasType`, the C-based type of all :pep:`695`-compliant
  type aliases and itself a valid type hint.
'''

# ....................{ FIXTURES                           }....................
def hints_nonpep695_meta() -> 'List[HintNonpepMetadata]':
    '''
    List of :pep:`695`-sorta-compliant **type hint metadata** (i.e.,
    :class:`beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintNonpepMetadata`
    instances describing test-specific :pep:`695`-sorta-compliant sample type
    hints with metadata generically leveraged by various PEP-agnostic unit
    tests).
    '''

    # ..................{ IMPORTS                            }..................
    # Defer fixture-specific imports.
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_12
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintNonpepMetadata,
        HintPithSatisfiedMetadata,
        HintPithUnsatisfiedMetadata,
    )

    # List of all PEP-noncompliant type hint metadata to be returned.
    hints_nonpep_meta = []

    # If the active Python interpreter targets Python < 3.12, this interpreter
    # fails to support PEP 695. In this case, return the empty list.
    if not IS_PYTHON_AT_LEAST_3_12:
        return hints_nonpep_meta
    # Else, the active Python interpreter targets Python >= 3.12 and thus
    # supports PEP 695.

    # ..................{ LISTS                              }..................
    # Defer version-specific imports.
    from beartype._cave._cavefast import HintPep695TypeAlias
    from beartype_test.a00_unit.data.pep.pep695.data_pep695hint import (
        AliasSimple)

    # Add PEP 695-specific (albeit technically PEP-noncompliant from the
    # beartype perspective) test type hints to this list.
    hints_nonpep_meta.extend((
        # Unsubscripted "typing.TypeAliasType" type of *ALL* PEP 695-compliant
        # type aliases. This type is a valid type and thus a valid hint, despite
        # defining a PEP-noncompliant "__parameters__" dunder attribute.
        HintNonpepMetadata(
            hint=HintPep695TypeAlias,
            piths_meta=(
                # Arbitrary PEP 695-compliant type alias.
                HintPithSatisfiedMetadata(AliasSimple),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    pith='And diamond-paved lustrous long arcades,',
                    # Match that the exception message raised for this pith
                    # contains...
                    exception_str_match_regexes=(
                        # The type *NOT* satisfied by this object.
                        r'\btyping\.TypeAliasType\b',
                    ),
                    # Match that the exception message raised for this pith does
                    # *NOT* contain...
                    exception_str_not_match_regexes=(
                        # A newline.
                        r'\n',
                        # A bullet delimiter.
                        r'\*',
                    ),
                ),
            ),
        ),
    ))

    # ..................{ RETURN                             }..................
    # Return this list of all PEP-noncompliant type hint metadata.
    return hints_nonpep_meta



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`- and :pep:`585`-compliant **dual type hint utilities**
(i.e., low-level callables generically applicable to both :pep:`484`- and
:pep:`585`-compliant type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep585Exception
from beartype._data.typing.datatypingport import (
    Hint,
    HintOrTupleHints,
)

# ....................{ GETTERS                            }....................
#FIXME: Unit test us up, please.
def get_hint_pep484585_arg(hint: Hint, exception_prefix: str) -> Hint:
    '''
    Sole child hint subscripting the passed :pep:`484`- or :pep:`585`-compliant
    parent hint if this hint is subscripted by exactly one child hint *or* raise
    an exception otherwise (i.e., if this parent hint is either unsubscripted
    *or* subscripted by two or more child hints).

    This getter is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as the implementation trivially reduces to
    an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Parent hint to be inspected.
    args_len : int
        Number of child hints expected to subscript this parent hint.
    exception_prefix : str
        Human-readable label prefixing the representation of this object in the
        exception message.

    Returns
    -------
    Hint
        Sole child hint subscripting this parent hint.

    Raises
    ------
    BeartypeDecorHintPep585Exception
        If this hint is subscripted by an unexpected number of child hints.

    See Also
    --------
    :func:`.get_hint_pep484585_args`
        Further details.
    '''

    # Defer to this lower-level getter.
    return get_hint_pep484585_args(  # type: ignore[return-value]
        hint=hint, args_len=1, exception_prefix=exception_prefix)


#FIXME: Unit test us up, please.
def get_hint_pep484585_args(
    hint: Hint, args_len: int, exception_prefix: str) -> HintOrTupleHints:
    '''
    Single child hint *or* tuple of all child hints subscripting the passed
    :pep:`484`- or :pep:`585`-compliant parent hint if this hint is subscripted
    by exactly the passed number of child hints *or* raise an exception
    otherwise.

    This getter returns either:

    * If ``args_len == 1``, the single child hint subscripting this parent hint
      as a convenience to the caller.
    * Else, the tuple of all child hints subscripting this parent hint.

    This getter is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as the implementation trivially reduces to
    an efficient one-liner.

    Caveats
    -------
    **This higher-level getter should always be called in lieu of directly
    accessing the low-level** ``__args__`` **dunder attribute,** which is
    typically *not* validated at runtime and thus should *not* be assumed to be
    sane. Although the :mod:`typing` module usually validates the arguments
    subscripting :pep:`484`-compliant type hints and thus the ``__args__``
    **dunder attribute at hint instantiation time, C-based CPython internals
    fail to similarly validate the arguments subscripting :pep:`585`-compliant
    type hints at any time:

    .. code-block:: python

        >>> import typing
        >>> typing.Type[str, bool]
        TypeError: Too many parameters for typing.Type; actual 2, expected 1
        >>> type[str, bool]
        type[str, bool]   # <-- when everything is okay, nothing is okay

    Parameters
    ----------
    hint : Hint
        Parent hint to be inspected.
    args_len : int
        Number of child hints expected to subscript this parent hint.
    exception_prefix : str
        Human-readable label prefixing the representation of this object in the
        exception message.

    Returns
    -------
    HintOrTupleHints
        Either the sole child hint *or* tuple of all child hints subscripting
        this parent hint.

    Raises
    ------
    BeartypeDecorHintPep585Exception
        If this hint is subscripted by an unexpected number of child hints.
    '''
    assert isinstance(args_len, int), f'{repr(args_len)} not integer.'
    assert args_len >= 1, f'{args_len} < 0.'

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepget import get_hint_pep_args

    # Tuple of all arguments subscripting this hint.
    hint_args = get_hint_pep_args(hint)

    # If this hint is *NOT* subscripted by the expected number of child type
    # hints...
    if len(hint_args) != args_len:
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        # Human-readable noun describing the grammatically correct plurality of
        # the number of expected child type hints. English! Why!?!?
        exception_noun = (
            'child type hint' if args_len == 1 else 'child type hints')

        # Raise an exception.
        raise BeartypeDecorHintPep585Exception(
            f'{exception_prefix}PEP 585 type hint {repr(hint)} '
            f'not subscripted (indexed) by {args_len} {exception_noun} (i.e., '
            f'subscripted by {len(hint_args)} != {args_len} child type hints).'
        )
    # Else, this hint is subscripted by the expected number of child type hints.

    # Return either...
    return (
        # If this hint is subscripted by only child hint, this child hint;
        hint_args[0]
        if args_len == 1 else
        # Else, this tuple of arguments as is.
        hint_args
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Decidedly Object-Oriented Runtime-checking (DOOR) subscripted type
hint classes** (i.e., :class:`beartype.door.TypeHint` subclasses implementing
support for :pep:`484`- and :pep:`585`-compliant subscripted type hints *not*
already matched by any more fine-grained :class:`beartype.door.TypeHint`
subclass).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.door._cls.doorsuper import TypeHint
from beartype.roar import BeartypeDoorPepArgsLenException
from beartype._data.hint.sign.datahintsignmap import (
    HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE)

# ....................{ SUBCLASSES                         }....................
class SubscriptedTypeHint(TypeHint):
    '''
    **Subscripted type hint wrapper** (i.e., high-level object encapsulating
    a low-level parent type hint satisfying various conditions).

    Notably, this wrapper wraps hints that both:

    * Are either :pep:`484`- or :pep:`585`-compliant.
    * Are subscripted (indexed) by a predetermined number of one or more
      low-level child type hints.
    * Originate from an **isinstanceable class** such that *all* objects
      satisfying this hint are instances of that class.
    '''

    # ..................{ PRIVATE ~ factories                }..................
    def _make_args(self) -> tuple:

        # Tuple of the zero or more low-level child type hints subscripting
        # (indexing) the low-level parent type hint wrapped by this wrapper.
        args = super()._make_args()

        # Argument length range (i.e., "range" object covering the minimum and
        # maximum number of child type hints that may subscript this low-level
        # parent type hint factory) if this factory has been associated with
        # such a range *OR* "None" otherwise (i.e., if this factory has *NOT*
        # been associated with such a range).
        args_len_range = HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE.get(
            self._hint_sign)  # type: ignore[arg-type]

        # If this factory has *NOT* been associated with such a range, raise an
        # exception. Note that this edge case should *NEVER occur. ----gulp----
        if args_len_range is None:  # pragma: no cover
            raise BeartypeDoorPepArgsLenException(  # pragma: no cover
                f'Type hint {repr(self._hint)} argument length range unknown.')
        # Else, this factory has been associated with such a range.

        #FIXME: Consider actually testing this. This *IS* technically
        #testable and should thus *NOT* be marked as "pragma: no cover".

        # If this hint was subscripted by an unexpected number of child hints...
        #
        # Note that this edge case commonly occurs with PEP 585-compliant type
        # hints (e.g., "list[str]"), which fail to validate their number of
        # child type hints: e.g.,
        #     >>> list[str, int]
        #     list[str, int]  # <-- wat
        if len(args) not in args_len_range:  # pragma: no cover
            #FIXME: This seems sensible, but currently provokes test failures.
            #Let's investigate further at a later time, please.
            # # If this hint was subscripted by *NO* parameters, comply with PEP
            # # 484 standards by silently pretending this hint was subscripted by
            # # the "typing.Any" fallback for all missing parameters.
            # if len(self._args) == 0:
            #     return (Any,)*self._args_len_expected

            # Exception message to be raised.
            exception_message = (
                f'PEP 585 type hint {repr(self._hint)} '
                f'not subscripted (indexed) by '
            )

            # Minimum and maximum number of arguments accepted by this factory.
            #
            # Note that the "stop" instance variable defined by "range" objects
            # is exclusive rather than inclusive. Substracting 1 from that
            # yields the inclusive maximum of this range.
            ARGS_LEN_MIN = args_len_range.start
            ARGS_LEN_MAX = args_len_range.stop - 1

            # If this factory accepts a constant (rather than variable) number
            # of child hints, raise a human-readable exception denoting this.
            if ARGS_LEN_MIN == ARGS_LEN_MAX:
                # Human-readable noun describing the grammatically correct
                # plurality of the number of expected child type hints. English!
                exception_noun = (
                    'child type hint' if len(args) == 1 else 'child type hints')

                # Append this number to this exception message.
                exception_message += (
                    f'{ARGS_LEN_MAX} {exception_noun} (i.e., '
                    f'subscripted by {len(args)} != '
                    f'{ARGS_LEN_MAX} child type hints).'
                )
            # Else, this factory accepts a variable number of child hints. Raise
            # a human-readable exception denoting this.
            else:
                # Append this number to this exception message.
                exception_message += (
                    f'[{ARGS_LEN_MIN}, {ARGS_LEN_MAX}] arguments (i.e., '
                    f'subscripted by {len(args)} child type hints).'
                )

            # Raise this exception.
            raise BeartypeDoorPepArgsLenException(exception_message)
        # Else, this hint was subscripted by the expected number of child hints.

        # Return these child hints.
        return args

    # ..................{ PRIVATE ~ testers                  }..................
    # Note that this redefinition of the superclass _is_equal() method is
    # technically unnecessary, as that method is already sufficiently
    # general-purpose to suffice for *ALL* possible subclasses (including this
    # subclass). Nonetheless, we wrote this method first. More importantly, this
    # method is *SUBSTANTIALLY* faster than the superclass method. Although
    # efficiency is typically *NOT* a pressing concern for the DOOR API,
    # discarding faster working code would be senseless.
    def _is_equal(self, other: TypeHint) -> bool:

        # If *ALL* of the child type hints subscripting both of these parent
        # type hints are ignorable, return true only if these parent type hints
        # both originate from the same isinstanceable class.
        if self._is_args_ignorable and other._is_args_ignorable:
            return self._origin == other._origin
        # Else, one or more of the child type hints subscripting either of these
        # parent type hints are unignorable.
        #
        # If either...
        elif (
            # These hints have differing signs *OR*...
            self._hint_sign is not other._hint_sign or
            # These hints have a differing number of child type hints...
            len(self._args_wrapped_tuple) != len(other._args_wrapped_tuple)
        ):
            # Then these hints are unequal.
            return False
        # Else, these hints share the same sign and number of child type hints.

        # Return true only if all child type hints of these hints are equal.
        return all(
            this_child == that_child
            #FIXME: Probably more efficient and maintainable to write this as:
            #    for this_child in self
            #    for that_child in other
            for this_child, that_child in zip(
                self._args_wrapped_tuple, other._args_wrapped_tuple)
        )
